# Copyright (C) 2006 Canonical Limited.
#
# Licence to be determined
#

"""A conflict finder for packages."""

import errno
import gc
import sha
import os.path
import re
import shlex
import shutil
import sys
import tempfile
import time
from unittest import TestSuite
import urllib

import apt_inst
import apt_pkg
from storm import properties, references

from findconflicts.tuned_gzip import *

def test_suite():
    """Return a test suite containing the find conflicts tests."""
    # local import as its not generally needed.
    import findconflicts.tests
    return findconflicts.tests.test_suite()


class ComponentGroupingHandler(object):
    """A conflict handler which groups the output by component."""

    def __init__(self, outputpath, pkg_components={}):
        """Construct a ComponentGroupingHandler.
        
        :param outputpath:  The path to output to. Must not end in /.
        :param pkg_components: A dictionary of
            package strings -> distro/components. i.e
            {'foo--1.2-i386':'warty/main'}.
            This can be set later by assigned to object.pkg_components.
        """
        self.components = {}
        self.pkg_components = pkg_components
        self.outputpath = outputpath
        self._workingpath = outputpath + '.new'

    def start(self):
        """Initialise the handler.

        This creates a temporary directory for outputting into.
        """
        shutil.rmtree(self._workingpath, ignore_errors=True)
        os.mkdir(self._workingpath)

    def conflict(self, pkg_left, pkg_right):
        """Report a conflict.

        This appends the conflict details to the component file for pkg_left.
        """
        pkg_left_str = str(pkg_left)
        if pkg_left_str not in self.pkg_components:
            components = 'unpublished'
        else:
            components = self.pkg_components[pkg_left_str]
        formatted_string = ("%s\n" %
                pkg_left.format_conflict(pkg_right, self.pkg_components))
        for component in components:
            if component not in self.components:
                output_path = self._workingpath + "/" + component + ".txt"
                if component != 'unpublished':
                    try:
                        os.mkdir(self._workingpath + "/" + os.path.dirname(component))
                    except OSError, e:
                        if e.errno == errno.EEXIST:
                            pass
                        else:
                            raise
                self.components[component] = open(output_path, 'wb')
            self.components[component].write(formatted_string)

    def finish(self):
        """Finish the report.

        This removes any existing output.old directory, renames the target
        output directory to .old, and moves the temporary directory into place,
        after closing all the files.
        """
        for file_obj in self.components.values():
            file_obj.close()
        shutil.rmtree(self.outputpath + '.old', ignore_errors=True)
        try:
            os.rename(self.outputpath, self.outputpath + '.old')
        except OSError, e:
            if e.errno == errno.ENOENT:
                pass
            else:
                raise
        os.rename(self._workingpath, self.outputpath)


class PackageDataSource(object):
    """A source of package metadata."""

    def __init__(self):
        # the default mirror to retrieve from
        self.set_mirror('http://archive.ubuntu.com/ubuntu/')
        # the default set of distributions to look at.
        # we should get this from launchpad
        gutsy_stock_components = {
            'main':['amd64', 'i386', 'powerpc', 'sparc'],
            'multiverse':['amd64', 'i386', 'powerpc', 'sparc'],
            'restricted':['amd64', 'i386', 'powerpc', 'sparc'],
            'universe':['amd64', 'i386', 'powerpc', 'sparc'],
            }
        # commercial repositories look like:
        commercial_components = {
            #'main':['amd64', 'i386', 'powerpc', 'sparc'],
            }
        # feisty == gutsy
        feisty_stock_components = gutsy_stock_components
        # edgy is currently identical to feisty
        edgy_stock_components = feisty_stock_components
        # dapper is currently identical to edgy
        dapper_stock_components = edgy_stock_components
        self.set_distributions({
            'dapper': dapper_stock_components,
            'dapper-backports': dapper_stock_components,
            'dapper-commercial': commercial_components,
            'dapper-security': dapper_stock_components,
            'dapper-updates': dapper_stock_components,
            'edgy': edgy_stock_components,
            'edgy-backports': edgy_stock_components,
            'edgy-commercial': commercial_components,
            'edgy-security': edgy_stock_components,
            'edgy-updates': edgy_stock_components,
            'feisty': feisty_stock_components,
            'feisty-backports': feisty_stock_components,
            'feisty-commercial': commercial_components,
            'feisty-security': feisty_stock_components,
            'feisty-updates': feisty_stock_components,
            'gutsy': gutsy_stock_components,
            'gutsy-backports': gutsy_stock_components,
            'gutsy-security': gutsy_stock_components,
            'gutsy-updates': gutsy_stock_components,
            })

    def _extract_package_identity(self, sections):
        """Extract the name version and arch triple from sections.

        This triple provides the unique identity of the binary.
        """
        info = {}
        info['name'] = sections['package']
        info['version'] = sections['version']
        info['arch'] = sections['architecture']
        return info

    def infoFromPackageUrl(self, url):
        """Return standard package information from a package at a url.
        
        The package is downloaded and analysed locally. If possible
        preinst diversions are detected, if that fails, the diversion list is
        set to ['UNKNOWN'] which cannot occur in regular packages.
        """
        package_file = urllib.urlopen(url)
        local_file = tempfile.TemporaryFile()
        try:
            # copy in 64k chunks
            data = package_file.read(64000)
            while len(data):
                local_file.write(data)
                data = package_file.read(64000)
            package_file.close()
            local_file.seek(0)
            control = apt_inst.debExtractControl(local_file)
            sections = apt_pkg.ParseSection(control)
            info = self._extract_package_identity(sections)
            info['controltext'] = control
            info['replaces'] = sections.get('replaces', '')
            info['conflicts'] = sections.get('conflicts', '')
            info['provides'] = sections.get('provides', '')
            info['files'] = []
            # find preinst diversions
            local_file.seek(0)
            preinst = apt_inst.debExtractControl(local_file, 'preinst')
            if preinst is not None:
                info['preinst_hash'] = sha.sha(preinst).hexdigest()
                try:
                    info['diversions'] = PreinstParser().extract_diversions(preinst)
                except ParseError:
                    info['diversions'] = ['UNKNOWN']
            # extract the archive and throw it away again.
            local_file.seek(0)
            def capture_filepaths(what, name, link, mode, uid, gid, size,
                mtime, major, minor):
                """A callback for debExtract to capture the file paths in use."""
                if what != "DIR":
                    info['files'].append(name)
            try:
                apt_inst.debExtract(local_file, capture_filepaths, "data.tar.gz")
            except SystemError, e:
                # pity there is not more precise an exception.
                if str(e) == "E:Cannot find chunk data.tar.gz":
                    local_file.seek(0)
                    apt_inst.debExtract(local_file, capture_filepaths, "data.tar.bz2")
                else:
                    raise
            return info
        finally:
            local_file.close()

    def get_packages(self, mirror=None, distribution=None, component=None, 
        arch=None):
        """Retrieve the parsed package list for mirror/distro/component/arch.
        
        :return: a dictionary mapping (name, version, arch) to url. e.g.
        {('atomix', '2.13.4-0ubuntu2', 'i386'):'http:/....'}
        """
        if mirror is None:
            mirror = self._mirror
        url = (mirror + "dists/" + distribution + "/" + component +
            "/binary-" + arch + "/Packages.gz")
        compressed_file = urllib.urlopen(url)
        gzip_file = GzipFile(fileobj=compressed_file)
        local_file = tempfile.TemporaryFile()
        result = {}
        try:
            # copy in 64k chunks
            try:
                data = gzip_file.read(64000)
            except Exception, e:
                print "error reading", url, e
                raise
            while len(data):
                local_file.write(data)
                data = gzip_file.read(64000)
            gzip_file.close()
            compressed_file.close()
            local_file.seek(0)
            # we now have a local temp file that is uncompresssed.
            parser = apt_pkg.ParseTagFile(local_file)
            while parser.Step() == 1:
                sections = parser.Section
                info = self._extract_package_identity(sections)
                url = mirror + sections['filename']
                result[(info['name'], info['version'], info['arch'])] = url
        finally:
            local_file.close()
        return result

    def get_all_packages(self, distributions=None):
        """Retrieve a dictionary of all the packages in distributions.

        :param distributions: a dictionary of distribution:components-dict,
        which then contains a list of architectures. e.g.
        {'dapper':{'main':['i386','sparc']}}
        :return: Two dictionaries.
            The first maps package tuples : package .deb urls.
            The second maps package tuples : list of distro/component.
        """
        urls = {}
        package_components = {}
        if distributions is None:
            distributions = self._distributions
        for distro, components in distributions.items():
            for component, arches in components.items():
                component_string = '%s/%s' % (distro, component)
                for arch in arches:
                    arch_packages = self.get_packages(
                        distribution=distro,
                        component=component,
                        arch=arch)
                    urls.update(arch_packages)
                    for package in arch_packages:
                        package_components.setdefault(
                            "%s_%s_%s" % package, set()).add(component_string)
                    
        return urls, package_components

    def set_distributions(self, distributions):
        """Set the distributions dictionary used for finding packages."""
        self._distributions = distributions

    def set_mirror(self, mirror):
        """Set the mirror to be used for retrieving package data."""
        self._mirror = mirror


class MissingConflictFinder(object):
    """The core of the system. This class finds missing conflicts."""

    def __init__(self):
        self.set_published_packages([])
        self.set_get_relationship(DebRelationship)

    def _conflicts_with(self, package_left, package_right):
        """Helper function - evaluate two packages for a one way conflict."""
        return self.get_relationship(package_left['conflicts']).matches(
            package_right['name'],
            package_right['version'],
            package_right['provides'])

    def concurrent_permitted(self, package_left, package_right):
        """Determine if the packages in package_pair can co-exist.

        This is the result of not(package_a.conflicts.matches(package_b) or
        package_b.conflicts.matches(package_a).
        """
        return not (self._conflicts_with(package_left, package_right) or
                    self._conflicts_with(package_right, package_left))
        
    def _replaces(self, package_left, package_right):
        """Helper function - evaluate two packages for a one way replaces."""
        return self.get_relationship(package_left['replaces']).matches(
            package_right['name'],
            package_right['version'])

    def concurrent_safe(self, package_left, package_right):
        """Determine if the packages in package_pair are 'safe' to co-exist

        This is the result of package_a.replaces.matches(package_b) or
        package_b.replaces.matches(package_a).
        """
        return (self._replaces(package_left, package_right) or
                self._replaces(package_right, package_left))
        
    def interesting_concurrent_installs(self, packages):
        """Calculate the interesting concurrent installs of packages.

        :param packages: a dictionary of (package,version,arch):{conflicts:'',
        replaces:''} data about packages. The conflicts nad replaces
        dictionaries are not used: this form of data is convient to other
        apis and is used for consistency here.
        :return: a set of the interesting concurrently installable pairs. Each
        element of the pair is a (package,version,arch) tuple.

        interesting concurrent installs are those which reflect bugs in the 
        current published set : we cannot correct missing conflicts in already
        'out there' package versions, but we can correct ones which occur
        between the current published package versions and any version we have
        encountered. We do this filtering in this routine as an efficiency 
        step : we can short circuit a lot of the combinatorial space for
        possible concurrent installs if we only generate those combinations
        containing published versions of any package.

        If no published packages are known, a full analysis is performed; this
        can be useful for reporting on the state of the world - how many 
        conflicts have happened in the past, how long it has taken to notice,
        what problems people may have previously have encountered, etc.

        In determining what is concurrently installable, this function only
        considers:
         * package name - only one instance of a package is installable.
         * architecture - only packages of the same architecture, or arch:all
           can be installed concurrently.
        This is so that the routine is single purpose and more useful : you
        can choose to analyse the return data to (for instance) see how many
        packages are concurrently installable in this manner, for hinting
        on where splits might make things simpler. See the concurrent_permitted
        routine to filter a set of concurrent installs down by
        conflicts/replaces.
        """
        # rearrange the packages list to be moe useful
        packages_by_name = {}
        # put all in to avoid special casing
        packages_by_arch = {'all':{}}

        for package in packages:
            name, version, arch = package
            packages_by_name.setdefault(name, set()).add(package)
            arch_names = packages_by_arch.setdefault(arch, {})
            arch_names.setdefault(name, set()).add(package)
        # pick up the package names that conflict
        conflicting_names = packages_by_name.keys()
        # short circuit for the common case of just one package
        if len(conflicting_names) == 1:
            return set()
        # generate all possible concurrent installs
        # we cant short circuit this by version because old versions on users
        # machines can still combine with new versions of other packages.
        # However we can reduce the size of the problem set by only examining 
        # one arch at a time - cross-arch combinations are not
        # current interesting, except for arch:all with arch:specific
        # we do have to process every arch because packages can choose to build
        # different binaries with different conflicts/replaces on different arch's.
        # TODO: one possibly interesting optimisation is to group any/all
        # identical conflicts/replaces/provides package versions
        # so that we only have to process the entire group once. This is
        # potentially a huge saving.
        result = set()
        for check_arch in packages_by_arch:
            # we now check each name present on the arch
            # but arch really means arch + all
            # TODO combine the two.
            if check_arch != 'all':
                # merge all into this arch
                packages_by_arch[check_arch].update(packages_by_arch['all'])
            concurrent_names = packages_by_arch[check_arch]
            possible_concurrent_installs = self._paired_permutations(*concurrent_names.values())
            result.update(possible_concurrent_installs)
        return result

    def _paired_permutations(self, *sets):
        """returns a list of the permutations of any two entries from sets.

        That is, if there are three sets with one member each, the return value
        is three pairs - (A1,B1), (A1, C1), (B1, C1). If the A set has 2 members
        then there are 5 permutations - the three above, and additionally (A2,B1),
        (A2, C1).
        """
        answer = set()
        pending = list(sets) 
        # pass once over the entire matrix.
        # roughly n^2
        # alternative: for unpublished entries, only generate permuations
        # with the published entries from the other sets, for published ones
        # generate all. This saves more as more unpublished entries are 
        # accrued. However, consider that two published packages with the 
        # same file and 20 versions is only 40 (2 * 20) variations anyway, so
        # it will be a while before this is a problem.
        # create our filter set.
        published = self.published_packages()
        if len(published) == 0:
            published = set()
            for a_set in pending:
                published.update(a_set)
        while len(pending):
            # grab a set to process, and remove it from the sets to combine with.
            current_set = pending.pop()
            # if this
            for element in current_set:
                # grab every other element of every other set in pending
                for other_set in pending:
                    for other_element in other_set:
                        if element in published or other_element in published:
                            answer.add(frozenset((element, other_element)))
        return answer

    def published_packages(self):
        """Return the set of (package, version, arch) that are published.

        You can set this via set_published_packages.

        :return: a set of package,version,arch tuples.
        """
        return self._published_packages

    def set_get_relationship(self, get_relationship):
        """Set the factory used to create relationships.

        This is used when evaluating conflicts: or replaces: clauses. It is
        primarily overridable to allow testing of the use of the relationship
        matcher, without the polynomial increase in needed tests if we were to
        test the composition of the two at once.
        """
        self.get_relationship = get_relationship

    def set_published_packages(self, packages):
        """Set the published packages collection.

        Published packages are used in determining what concurrency situations
        are interesting.
        """
        self._published_packages = set(packages)


class BinaryPackageFiles(object):
    """The files that are installed by a binary package. 
    
    Each file belongs to some number of packages, and likewise, each package
    has some number of files.

    Adding a new set of files will return all the packages across all the files
    that overlap.

    The relationship is bidirectional: its possible to ask for the files for a
    package, the packages that potentially collide on files for a given package,
    or the packages that provide a file.
    """

    def __init__(self):
        self._files = {}

    def add_files(self, files, package):
        """Add files to the collection, associating them with package."""
        result = {}
        for file in files:
            file_packages = self._files.setdefault(file, set())
            # add package
            file_packages.add(package)
            different_packages = filter(lambda pkg:pkg[0] != package[0], file_packages)
            different_packages.append(package)
            if len(different_packages) > 1:
                files_conflicting = result.setdefault(frozenset(different_packages), [])
                files_conflicting.append(file)
        return result


class DebRelationship(object):
    """A relation between debian packages."""

    def __init__(self, relationship):
        """Construct a DebRelation from the relationship in text form.

        The standard debian operators - ',', '<<', '<=', '=', '>=', '>>' are 
        supported, as are the legacy '<' and '>' operators.
        """
        self._rule = apt_pkg.ParseDepends(relationship)

    def matches(self, matchname, matchversion, provides=''):
        """Evaluate whether package is matched by this relationship."""
        # test package against replaces
        # FIXME: this should whole function should be a call into apt_pkg
        provides = set([
            element[0][0] for element in apt_pkg.ParseDepends(provides)])
        for or_rule in self._rule:
            # each of these is an and rule in theory, but for both
            # conflicts and replaces, and never makes sense.
            for packagename, version, operator in or_rule:
                if packagename != matchname:
                    # not for this package
                    # but might be for a virtual package
                    if version == '' and packagename in provides:
                        return True
                    else:
                        continue
                if version == '':
                    return True
                versiondiff = apt_pkg.VersionCompare(matchversion, version)
                if operator == '<<':
                    return versiondiff < 0
                elif operator in ('<=', '<'):
                    return versiondiff <= 0
                elif operator == '=':
                    return versiondiff == 0
                elif operator in ('>=', '>'):
                    return versiondiff >= 0
                elif operator == '>>':
                    return versiondiff > 0
                else:
                    raise RuntimeError('unrecognised operator %r' % operator)
            # rule
            #if rule[0][0] == package[0]:
            #    return True
        return False


class ArchiveDatabase(object):
    """A place to put routines referring to the general idea of the Archive."""

    @staticmethod
    def initialize(uri):
        """Initialise an Archive database on uri."""
        from storm.locals import create_database, Store
        database = create_database(uri)
        store = Store(database)
        # ---- architecture ----
        store.execute(
            "CREATE TABLE architecture "
            "(id INTEGER PRIMARY KEY,"
            " arch VARCHAR)")
        store.execute("CREATE UNIQUE INDEX _unique_arch ON architecture (arch)")
        # ---- packagename ----
        store.execute(
            "CREATE TABLE packagename "
            "(id INTEGER PRIMARY KEY,"
            " name VARCHAR)")
        store.execute("CREATE UNIQUE INDEX _unique_packagename ON packagename (name)")
        # ---- packageversion ----
        store.execute(
            "CREATE TABLE packageversion "
            "(id INTEGER PRIMARY KEY,"
            " name_id INTEGER,"
            " version VARCHAR,"
            " arch_id INTEGER,"
            " controltext VARCHAR,"
            " analyzed_level INTEGER,"
            " preinst_hash VARCHAR)")
        store.execute("CREATE UNIQUE INDEX _unique_version ON packageversion "
            "(name_id, version, arch_id)")
        store.execute("CREATE INDEX _packageversion_analyzed on packageversion "
            "(analyzed_level)")
        # ---- filepath ----
        store.execute(
            "CREATE TABLE filepath "
            "(id INTEGER PRIMARY KEY,"
            " path VARCHAR)")
        store.execute("CREATE UNIQUE INDEX _unique_filepath ON filepath "
            "(path)")
        # ---- binpackagefiles ----
        store.execute(
            "CREATE TABLE binpackagefiles "
            "(package_id INTEGER,"
            " file_id INTEGER)")
        store.execute("CREATE UNIQUE INDEX _unique_packagefile ON binpackagefiles "
            "(package_id, file_id)")
        store.execute("CREATE INDEX _bpf_fileid on binpackagefiles (file_id)")
        # ---- diversion ----
        store.execute(
            "CREATE TABLE diversion "
            "(package_id INTEGER,"
            " file_id INTEGER)")
        store.execute("CREATE UNIQUE INDEX _unique_diversion ON diversion"
            "(package_id, file_id)")
        # --------------------
        store.commit()
        return store

    @staticmethod
    def open_or_create():
        """Open or create if needed the default store 'findconflicts.db'."""
        from storm.locals import create_database, Store
        from storm.exceptions import OperationalError
        uri = 'sqlite:findconflicts.db'
        database = create_database(uri)
        store = Store(database)
        try:
            list(store.find(PackageName))
            return store
        except OperationalError, e:
            if str(e) != 'no such table: packagename':
                # unexpected error - bail
                raise
            # expected error in a new db - make a new one
            del database
            del store
            return ArchiveDatabase.initialize(uri)


class Filepath(object):
    """A filepath contained in a binary package."""

    __table__ = "filepath", ("id", "path")
    id = properties.Int()
    path = properties.Str()

    def __init__(self, path):
        """Create a Filepath.

        :param path: The path to be recorded.
        """
        self.path = path

    @staticmethod
    def Ensure(path, store):
        """Ensure that there is a path called path in store store and return it.

        :param path: The string for the path.
        :param store: The store to use.
        :return: A Filepath.
        """
        try:
            return list(store.find(Filepath, path=path))[0]
        except IndexError:
            result = Filepath(path)
            store.add(result)
            return result

    def __eq__(self, other):
        return other is not None and other.path == self.path

    def __lt__(self, other):
        return self.path < other.path

    def __repr__(self):
        return "<findconflicts.Filepath at %d %s>" % (
            id(self), self.path)

    def __str__(self):
        return self.path


class PackageArchitecture(object):
    """The architecture of a binary package."""

    __table__ = "architecture", ("id", "arch")
    id = properties.Int()
    arch = properties.Str()

    def __init__(self, arch):
        """Create a PackageArchitecture.

        :param arch: The string representing the architecture. I.e. 'all'.
        """
        self.arch = arch

    def __eq__(self, other):
        return other is not None and other.arch == self.arch

    @staticmethod
    def Ensure(arch, store):
        """Ensure that there is a arch 'name' in store store and return it.

        :param arch: The string for the name.
        :param store: The store to use.
        :return: A PackageArchitecture.
        """
        try:
            return list(store.find(PackageArchitecture, arch=arch))[0]
        except IndexError:
            result = PackageArchitecture(arch)
            store.add(result)
            return result

    def __str__(self):
        return self.arch


class PackageFile(object):
    """An occurance of a file in a specific package.
    
    This class should not be used at all, it only exists to 
    facilitate describing the object links to STorm.
    """

    __table__ = "binpackagefiles", ("package_id", "file_id")
    package_id = properties.Int()
    file_id = properties.Int()


class PackageName(object):
    """The name of a binary package."""

    __table__ = "packagename", ("id", "name")
    id = properties.Int()
    name = properties.Str()

    def __init__(self, name):
        """Create a PackageName.

        :param name: The string representing the name. I.e. 'atomix'.
        """
        self.name = name

    def __eq__(self, other):
        return other is not None and other.name == self.name

    @staticmethod
    def Ensure(name, store):
        """Ensure that there is a name 'name' in store store and return it.

        :param name: The string for the name.
        :param store: The store to use.
        :return: A PackageName.
        """
        try:
            return list(store.find(PackageName, name=name))[0]
        except IndexError:
            result = PackageName(name)
            store.add(result)
            return result

    def __str__(self):
        return self.name


class PackageVersion(object):
    """The name of a binary package."""

    __table__ = "packageversion", "id"
    id = properties.Int()
    controltext = properties.Str()
    name_id = properties.Int()
    name = references.Reference(name_id, PackageName.id)
    version = properties.Str()
    arch_id = properties.Int()
    arch = references.Reference(arch_id, PackageArchitecture.id)
    analyzed_level = properties.Int()
    files = references.ReferenceSet(id, PackageFile.package_id,
        PackageFile.file_id, Filepath.id)
    preinst_hash = properties.Str()

    def __init__(self, packagename, version, packagearch, controltext,
        analyzed_level=0, diversions=None, preinst_hash=None):
        """Create a PackageVersion.

        :param packagename: The PackageName for the version.
        :param version: A string representing the version.
        :param packagearch: The PackageArchitecture for the version.
        :param controltext: The text of the control document for the version.
            This is used to generate the conflicts/replaces etc sections on
            demand.
        :param analyzed_level: The amount of analysis performed on this package.
            Current values: 0 - None, 1 - simple conflicts/replaces/provides.
        :param diversions: An iterable of the diversions introduced by this
            package.
        """
        self.controltext = controltext
        self.name = packagename
        self.version = version
        self.arch = packagearch
        self.analyzed_level = analyzed_level
        self.preinst_hash = preinst_hash or "da39a3ee5e6b4b0d3255bfef95601890afd80709"

    def __eq__(self, other):
        return (other is not None and 
            other.name == self.name and
            other.version == self.version and
            other.arch == self.arch and
            other.controltext == self.controltext and
            other.analyzed_level == self.analyzed_level and
            other.preinst_hash == self.preinst_hash)

    def __load__(self):
        """Set non-storm managed attributes on load."""
        store = getattr(self, '__object_info')['store']
        result = store.execute(
                    "SELECT path FROM packageversion, diversion, filepath "
                    "WHERE packageversion.id=? AND "
                    "packageversion.id=diversion.package_id AND "
                    "diversion.file_id=filepath.id", [self.id])
        self.diversions = set(str(row[0]) for row in result)

    def analyse(self, finder):
        """Analyse this package against the stores collection.

        If no conflicts are found, the packages analyzed_level is set to the
        current analysis level (1). This is set to mark the analysis as both
        complete and needing no further attention.

        :param finder: A MissingConflictFinder to use in the analysis.
        :return: A list of conflicts.
        """
        possible_conflicts = self.find_unsafely_colocated_packages()
        conflicting_packages = []
        for conflict_package in possible_conflicts:
            if self.conflicts_against_package(conflict_package, finder):
                conflicting_packages.append(conflict_package)
        if not conflicting_packages:
            self.analyzed_level = 1
        return conflicting_packages

    @staticmethod
    def Analyse_ids(package_ids, store, finder, pkg_components, log, handler=None):
        """Analyse package_ids in store, logging to log.

        :param package_ids: The database ids of the package versions to analyse.
            Must be iterable.
        :param store: The store to perform database queries on.
        :param finder: a MissingConflictFinder to use.
        :param pkg_components: A dict mapping package strings to a list of 
            distro/components.
        :param log: A stream to write results to.
        :param handler: A ConflictHandler.
        :return: Nothing.
        """
        errors_found = False
        total = len(package_ids)
        if handler is not None:
            handler.start()
        for index, package_id in enumerate(package_ids):
            start = time.time()
            package = store.find(PackageVersion, id=package_id).one()
            conflicting_packages = package.analyse(finder)
            if not conflicting_packages:
                # dont use the regular commit as it exhibits massive slowdowns
                # after thousands of commits.
                store.flush()
                store._connection.commit()
                #store.commit()
            else:
                errors_found = True
                for conflict_package in conflicting_packages:
                    if handler is not None:
                        handler.conflict(package, conflict_package)
            log.write("Finished %s in %.3f seconds [%d/%d]\n" % (
                package, time.time() - start, index, total))
            log.flush()
            #if not index % 500:
            #    # stop storm going slower.
            #    gc.collect()
        if not errors_found:
            log.write("no errors (checked %d packages).\n" % total)
            log.flush()
        if handler is not None:
            handler.finish()

    def conflicts_against_package(self, packageversion, finder):
        """Determine if this package and packageversion are safe together.

        :param packageversion: A packageversion to check.
        :param finder: A MissingConflictFinder to perform the check with.
        :return: True if there is a conflict that the packaging system cannot
            predict (i.e. a missing replaces or conflicts clause).
        """
        current_info = self.get_info()
        conflict_info = packageversion.get_info()
        if (finder.concurrent_permitted(current_info, conflict_info) and not
            finder.concurrent_safe(current_info, conflict_info)):
            return True
        return False

    @staticmethod
    def Find(package_tuple, store):
        """Try to find package package_tuple in store."""
        results = store.find(
            (PackageName, PackageVersion, PackageArchitecture),
            PackageName.id == PackageVersion.name_id,
            PackageArchitecture.id == PackageVersion.arch_id,
            PackageName.name == package_tuple[0],
            PackageVersion.version == package_tuple[1],
            PackageArchitecture.arch == package_tuple[2],
            ).one()
        if not results:
            raise KeyError
        return results[1]
        
#        result = store.find(PackageVersion,
#            "packagename.name=%s and packageversion.version=%s and packagearch.arch=%s" % package_tuple).one()
#            name = PackageName(package_tuple[0]),
#            version = package_tuple[1],
#            arch = PackageArchitecture(package_tuple[2])).one()
        if result is None:
            raise KeyError(package_tuple)
        return result

    def find_unsafely_colocated_packages(self):
        """Return a list of packages which install files this package does.

        Packages that are returned are soley those that the package manager
        *might* allow to be coinstalled - packages with the same name, or on
        architectures that are mutually incompatible (i.e. ppc vs i386) are
        ignored - and which are not known to be ok to coinstall.

        Files that are diverted by either end of a pair of packages are
        considered safe.

        This function deliberately does not perform advanced filtering,
        to allow callers to perform their own analysis.
        """
        # peek under the hood to get the store instance to use.
        store = getattr(self, '__object_info')['store']
        all = PackageArchitecture.Ensure('all', store)
        store.flush()
        all_id = all.id
        if self.arch.arch != 'all':
            candidate_arches = [self.arch_id, all_id]
        else:
            candidate_arches = [arch.id for arch in iter(store.find(PackageArchitecture))]
        # search for package tuples: 
        query = ("SELECT DISTINCT packageversion.id FROM "
            "packageversion, binpackagefiles, binpackagefiles as bpf2 "
            "WHERE bpf2.package_id=%d AND " #the bpf for this package
            " bpf2.file_id=binpackagefiles.file_id AND " # and all bpf that intersect
            " bpf2.file_id NOT IN (SELECT file_id FROM diversion WHERE package_id in (%d,binpackagefiles.package_id)) AND" # filter out diversions - by knocking out files from this package that are diverted by it or the other package.
            " packageversion.arch_id IN "
            "(" % (self.id, self.id) + "?, " * len(candidate_arches) + ") AND "
            "packageversion.id = binpackagefiles.package_id AND " # limit to the matching packageversion
            "packageversion.name_id != %d" % self.name_id #exclude the same named packge
            )
        candidates = [result[0] for result in store.execute(query, candidate_arches).get_all()]
        return sorted(store.find(PackageVersion,
            PackageVersion.id.is_in(candidates)), key=lambda x:str(x))

    def find_common_files(self, other):
        """Find the files in common with other - useful for debugging.

        :param other: Another PackageVersion.
        """
        # peek under the hood to get the store instance to use.
        store = getattr(self, '__object_info')['store']
        common_ids = [result[0] for result in store.execute(
            "SELECT DISTINCT binpackagefiles.file_id FROM binpackagefiles, "
            "binpackagefiles as bpf2 "
            "WHERE binpackagefiles.package_id = %d AND "
            "bpf2.package_id = %d AND binpackagefiles.file_id = bpf2.file_id"
            % (self.id, other.id)
            ).get_all()]
        return sorted(store.find(Filepath, Filepath.id.is_in(common_ids)))

    @staticmethod
    def PackageTuplesToUnicode(package_tuples):
        """Convery package_tuples into a unicode string list for use in queries.

        :param package_tuples: An iterable of (name, version, arch) tuples.
        """
        return [unicode("%s_%s_%s") % one_tuple for one_tuple in package_tuples]

    @staticmethod
    def Find_new_versions(package_tuples, store):
        """Return the set subtraction of package_tuples and the store."""
        # We select all the packages that do exist from the db, and then do
        # a set subtraction within python. If the DB layer starts to error
        # due to the size of the IN list, we can just do it in chunks of X
        # thousand.
        package_strings = PackageVersion.PackageTuplesToUnicode(package_tuples) 
        existing_versions = set(store.execute("SELECT packagename.name, "
            " packageversion.version, architecture.arch FROM "
            "packageversion, packagename, architecture "
            "WHERE packagename.id=packageversion.name_id and "
            "packageversion.arch_id=architecture.id AND "
            "packagename.name||'_'||packageversion.version||'_'||architecture.arch IN (%s)" % ("?," * len(package_strings)),
            package_strings).get_all())
        return set(package_tuples) - existing_versions

    def format_components(self, pkg_string, pkg_components):
        """Format a pkg string with the distro components it is in.

        :param pkg_string: a package version string.
        :param pkg_components: A dict mapping package strings to a list of 
            distro/components.
        :return: A string for display.
        """
        components = pkg_components.get(pkg_string, ['not published'])
        return pkg_string + "[" + ", ".join(components) + "]"

    def format_conflict(self, conflict_package, pkg_components):
        """Return a string describing the conflict with conflict_package.

        :param conflict_package: A package that a conflict has been reported
            against.
        :param pkg_components: A dict mapping package strings to a list of 
            distro/components.
        :return: A string describing the conflict.
        """
        self_str = self.format_components(str(self), pkg_components)
        conflict_str = self.format_components(str(conflict_package),
            pkg_components)
            
        return "%s, %s %s" % (self_str, conflict_str,
            [str(f) for f in self.find_common_files(conflict_package)])

    def get_info(self):
        """Return a dict with key information about this package.

        This is in the format expected by the MissingConflictFinder.
        :return: A dict with the keys name, version, arch, conflicts, provides 
            and replaces.
        """
        result = {}
        result['name'] = self.name.name
        result['version'] = self.version
        result['arch'] = self.arch.arch
        sections = apt_pkg.ParseSection(self.controltext)
        result['conflicts'] = sections.get('conflicts', '')
        result['provides'] = sections.get('provides', '')
        result['replaces'] = sections.get('replaces', '')
        return result

    @staticmethod
    def InsertFromInfo(info, store):
        """Insert the package details from 'info' into the store 'store'.

        :param info: A dictionary as returned by infoFromPackageUrl.
        """
        arch = PackageArchitecture.Ensure(info['arch'], store)
        name = PackageName.Ensure(info['name'], store)
        version = PackageVersion(
            name, info['version'], arch, info['controltext'])
        if "preinst_hash" in info:
            version.preinst_hash = info['preinst_hash']
        store.add(version)
        # we need version to have an id.
        store.flush()
        # 1) determine file paths to add - both files and diversions
        allfiles = set(info['files'])
        allfiles.update(set(info.get('diversions', [])))
        existing_files = set(map(str, iter(store.find(Filepath, Filepath.path.is_in(allfiles)))))
        new_files = allfiles - existing_files
        # 2) add them
        for path in new_files:
            sqlresult = store.execute("INSERT INTO filepath (path) VALUES (?)",
                (path,)).get_all()
            assert sqlresult == []
        # 3) insert all the paths into this table.
        query = (
            "INSERT INTO binpackagefiles (package_id, file_id) SELECT %d,"
            " filepath.id FROM filepath WHERE filepath.path in (" % version.id
            + "?, " * len(info['files']) + ")" )
        sqlresult = store.execute(query, info['files']).get_all()
        assert sqlresult == []
        # insert diversions
        if info.get('diversions', []):
            diversions = list(set(info['diversions']))
            query = (
            "INSERT INTO diversion (package_id, file_id) SELECT %d,"
            " filepath.id FROM filepath WHERE filepath.path in (" % version.id
            + "?, " * len(diversions) + ")" )
            sqlresult = store.execute(query, diversions).get_all()
            assert sqlresult == []
            version.diversions = set(diversions)
        return version

    def __repr__(self):
        return "<findconflicts.PackageVersion at %d %s_%s_%s>" % (
            id(self), self.name, self.version, self.arch)

    def __str__(self):
        return "%s_%s_%s" % (self.name, self.version, self.arch)

    @staticmethod
    def To_analyze(store, published_packages=None):
        """Return the packageversion_ids that need further analysis.

        :param store: A Storm store to use to retrieve the ids.
        :param published_packages: A subset of packages to examine.
        :return: An iterable of ids.
        """
        if not published_packages:
            params = None
            query = "select id from packageversion where analyzed_level = 0"
        else:
            params = PackageVersion.PackageTuplesToUnicode(published_packages)
            query = ("SELECT packageversion.id "
                "FROM packageversion, packagename, architecture "
                "WHERE packageversion.analyzed_level = 0 AND "
                "packagename.id=packageversion.name_id and "
                "packageversion.arch_id=architecture.id AND "
                "packagename.name||'_'||packageversion.version||'_'||architecture.arch IN (%s)" % ("?," * len(params)))
        return [result[0] for result in store.execute(query, params)]

Filepath.binpackages = references.ReferenceSet(Filepath.id, PackageFile.file_id,
        PackageFile.package_id, PackageVersion.id)
PackageName.versions = references.ReferenceSet(PackageName.id, PackageVersion.name_id)


class ParseError(Exception):
    """A parsing error occured."""


class PreinstParser(object):
    """A parser for debian preinst scripts."""

    def extract_diversions(self, bytes):
        """Extract diversion information from the preinst file 'bytes'.

        :param bytes: A debian preinst script.
        :return a list of diversions made by the script.
        :raise: When an unparseable script is given, ParseError is raised.
        """
        lines = bytes.split('\n')
        if lines[0].startswith("#!/bin/sh"):
            lang="shell"
        elif lines[0].startswith("#! /bin/sh"):
            lang="shell"
        elif lines[0].startswith("#!/bin/bash"):
            lang="shell"
        elif lines[0].startswith("#! /bin/bash"):
            lang="shell"
        else:
            raise ParseError("unknown language")
        # assume shell for now.
        # combine \ extended lines.
        normallines = []
        continuation = False
        for line in lines[1:]:
            if continuation:
                normallines[-1] = normallines[-1][:-1] + line
            else:
                normallines.append(line)
            # TODO: ignore \ at the end of comments.
            continuation = line.endswith('\\')
        result = []
        in_string = False
        for line in normallines:
            # strip out comments.
            try:
                tokens = shlex.split(line, True)
            except ValueError, e:
                if e.args == ('No closing quotation', ):
                    # tried to split a multiline string: ignore text until
                    # the next one (which will be end of the multiline string
                    in_string = not in_string
                    # and skip this line regardless.
                    continue
                else:
                    print line
                    raise
            if in_string:
                continue
            # rejoin the normalised line for regex processing.
            line = ' '.join(tokens)
            matchobj = re.search(
                ".*dpkg-divert\s+--package\s+\S+\s+(?:(?:(?:--add)|(?:--rename)|(?:--divert\s+\S+))\s+){3}(\S+)$", line)
            if matchobj:
                result.append(matchobj.groups()[0])

        return result


apt_pkg.InitSystem()
