# Copyright (C) 2006 Canonical Limited.
#
# Licence to be determined
#

"""Tests for finding absent conflicts."""

from unittest import TestCase

from findconflicts import (
    DebRelationship,
    MissingConflictFinder,
    PackageDataSource,
    )


class TestMissingConflictFinder(TestCase):

    def test_construct(self):
        # default constructor of a MissingConflictFinder has no published
        # packages.
        finder = MissingConflictFinder()
        self.assertEqual(set(), finder.published_packages())

    def test_set_published_package(self):
        # we can set new published_packages
        finder = MissingConflictFinder()
        published_packages = set([
            ('a', '1.2.1', 'i386'),
            ('a', '1.4.3', 'sparc'),
            ])
        finder.set_published_packages(published_packages)
        self.assertEqual(published_packages, finder.published_packages())
        # once set it must be independent
        published_packages.add(('b', '1.2.3', 'i386'))
        self.assertEqual(
            set([('b', '1.2.3', 'i386')]),
            published_packages - finder.published_packages())

    def test__paired_permutations(self):
        # the _paired_permutations utility routine filters out by the
        # published_packages set.
        # check some common cases for paired_permutations
        # first with no published packages.
        finder = MissingConflictFinder()
        self.assertEqual(set(), finder._paired_permutations())
        self.assertEqual(set(), finder._paired_permutations(set()))
        self.assertEqual(set(), finder._paired_permutations(set(['A1'])))
        self.assertEqual(set(), finder._paired_permutations(set(['A1']), set()))
        self.assertEqual(set(), finder._paired_permutations(set(), set(['B1'])))
        self.assertEqual(set((frozenset(['A1', 'B1']), )),
            finder._paired_permutations(
                set(['A1']),
                set(['B1'])
                )
            )
        self.assertEqual(
            set((
                frozenset(['A1', 'B1']), 
                frozenset(['A2', 'B1']), 
                )),
            finder._paired_permutations(
                set(['A1', 'A2']),
                set(['B1']),
                )
            )
        self.assertEqual(
            set((
                frozenset(['A1', 'B1']), 
                frozenset(['A2', 'B1']), 
                frozenset(['A1', 'C1']), 
                frozenset(['A2', 'C1']), 
                frozenset(['B1', 'C1']), 
                )),
            finder._paired_permutations(
                set(['A1', 'A2']),
                set(['B1']),
                set(['C1']),
                )
            )
        self.assertEqual(
            set((
                frozenset(['A1', 'B1']), 
                frozenset(['A2', 'B1']), 
                frozenset(['A1', 'C1']), 
                frozenset(['A2', 'C1']), 
                frozenset(['B1', 'C1']), 
                frozenset(['A1', 'C2']), 
                frozenset(['A2', 'C2']), 
                frozenset(['B1', 'C2']), 
                )),
            finder._paired_permutations(
                set(['A1', 'A2']),
                set(['B1']),
                set(['C1', 'C2']),
                )
            )
        # now with a published package set - representing one package at -2
        # and one package with two versions published -1 and -2 (perhaps in 
        # two distributions
        finder.set_published_packages(set(['A2', 'C1', 'C2']))
        self.assertEqual(set(),
            finder._paired_permutations(
                set(['A1']),
                set(['B1'])
                )
            )
        self.assertEqual(
            set((
                frozenset(['A2', 'B1']),
                )),
            finder._paired_permutations(
                set(['A1', 'A2']),
                set(['B1']),
                )
            )
        self.assertEqual(
            set((
                frozenset(['A2', 'B1']),
                frozenset(['A1', 'C1']),
                frozenset(['A2', 'C1']),
                frozenset(['B1', 'C1']),
                )),
            finder._paired_permutations(
                set(['A1', 'A2']),
                set(['B1']),
                set(['C1']),
                )
            )
        self.assertEqual(
            set((
                frozenset(['A2', 'B1']),
                frozenset(['A1', 'C1']),
                frozenset(['A2', 'C1']),
                frozenset(['B1', 'C1']),
                frozenset(['A1', 'C2']),
                frozenset(['A2', 'C2']),
                frozenset(['B1', 'C2']),
                )),
            finder._paired_permutations(
                set(['A1', 'A2']),
                set(['B1']),
                set(['C1', 'C2']),
                )
            )


class TestInterestingConcurrentInstallsNoPublishedPackages(TestCase):
    """Tests for the interesting_concurrent_installs feature.

    a Missing conflict finder can report on 'interesting' possible
    concurrent installs. With no published_package set this should
    report all possible combinations of packages given to it.
    
    With a published set it should only classify interesting possible
    concurrent installs as those involving one of the currently published
    packages in it.
    """
        
    def setUp(self):
        """All tests here need a finder."""
        super(TestInterestingConcurrentInstallsNoPublishedPackages, self).setUp()
        self.finder = MissingConflictFinder()
        self.null_replace_conflicts = {
            'replaces':'',
            'conflicts':'',
            }

    def test_calculate_interesting_possible_concurrent_installs(self):
        # nothing in, nothing out.
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs({}))

    def test_same_name_is_not_concurrently_installable(self):
        # the same package always counts as not concurrently installable.
        # regardless of arch, version etc.
        packages = {
            ('atomix', '1.2.1-1', 'all'):{},
            ('atomix', '1.2.1-1', 'i386'):{},
            ('atomix', '1:1.2.1-1', 'sparc'):{},
            }
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))

    def test_two_package_names_no_conflict_replace(self):
        # when two separate packages on the same arch
        # do not conflict or replace,
        # they are concurrently installable and we get all combinations of 
        # two packages back..
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'i386')
        package_b_2 = ('b', '2-2', 'i386')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set([frozenset([package_a_1, package_b_1])] ),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_2, package_b_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_1, package_b_2]),
                frozenset([package_a_2, package_b_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))

    def test_two_package_names_one_all_no_conflict_replace(self):
        # when two separate packages on the same arch
        # do not conflict or replace,
        # they are concurrently installable. 'all' counts as the
        # same arch for the purposes of this test, so its nearly a dup
        # of the test_two_package_names_no_conflict_replace test.
        # but there isn't enough actual duplicate intent to 
        # factor out yet usefully.
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'all')
        package_b_2 = ('b', '2-2', 'all')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set([frozenset([package_a_1, package_b_1])] ),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_2, package_b_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_1, package_b_2]),
                frozenset([package_a_2, package_b_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))

    def test_different_arches_not_concurrent(self):
        # this tests that that separate arches do not impact each other 
        # for concurrency.
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'sparc')
        package_b_2 = ('b', '2-2', 'sparc')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))

    def test_three_package_names_no_conflict_replace(self):
        # when three separate packages on the same arch
        # do not conflict or replace,
        # they are concurrently installable
        # but we dont care that its a group, we want to know about
        # the individual pairs that can occur
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'i386')
        package_b_2 = ('b', '2-2', 'i386')
        package_c_1 = ('c', '2-1', 'i386')
        package_c_2 = ('c', '2-2', 'i386')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            package_c_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_c_1]),
                frozenset([package_b_1, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a present as well.
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_c_1]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_c_1]),
                frozenset([package_b_1, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_b_2]),
                frozenset([package_a_1, package_c_1]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_a_2, package_c_1]),
                frozenset([package_b_1, package_c_1]),
                frozenset([package_b_2, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # and finally with a second c
        packages[package_c_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_b_2]),
                frozenset([package_a_1, package_c_1]),
                frozenset([package_a_1, package_c_2]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_a_2, package_c_1]),
                frozenset([package_a_2, package_c_2]),
                frozenset([package_b_1, package_c_1]),
                frozenset([package_b_2, package_c_1]),
                frozenset([package_b_1, package_c_2]),
                frozenset([package_b_2, package_c_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))

    def test_three_package_names_one_arch_all_no_conflict_replace(self):
        # when three separate packages 
        # do not explicitly conflict or replace,
        # they are concurrently installable, even if one is on arch: all
        # but we dont care that its a group, we want to know about
        # the individual permutations that can occur.
        package_a_1 = ('a', '1-1', 'all')
        package_a_2 = ('a', '1-2', 'all')
        package_b_1 = ('b', '2-1', 'i386')
        package_b_2 = ('b', '2-2', 'i386')
        package_c_1 = ('c', '2-1', 'i386')
        package_c_2 = ('c', '2-2', 'i386')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            package_c_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_c_1]),
                frozenset([package_b_1, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a present as well.
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_c_1]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_c_1]),
                frozenset([package_b_1, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_b_2]),
                frozenset([package_a_1, package_c_1]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_a_2, package_c_1]),
                frozenset([package_b_1, package_c_1]),
                frozenset([package_b_2, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # and finally with a second c
        packages[package_c_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_b_2]),
                frozenset([package_a_1, package_c_1]),
                frozenset([package_a_1, package_c_2]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_a_2, package_c_1]),
                frozenset([package_a_2, package_c_2]),
                frozenset([package_b_1, package_c_1]),
                frozenset([package_b_2, package_c_1]),
                frozenset([package_b_1, package_c_2]),
                frozenset([package_b_2, package_c_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))

    def test_three_package_names_two_arch_one_all_no_conflict_replace(self):
        # when three separate packages 
        # do not explicitly conflict or replace, and one is arch X, one is
        # arch Y, one is arch: all, the X and all, and Y and all are
        # concurrently installable.
        # they are concurrently installable, even if one is on arch: all
        # but we dont care that its a group, we want to know about
        # the individual errors that can occur
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'all')
        package_b_2 = ('b', '2-2', 'all')
        package_c_1 = ('c', '2-1', 'sparc')
        package_c_2 = ('c', '2-2', 'sparc')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            package_c_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_b_1, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a present as well.
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_b_1, package_c_1]),
                frozenset([package_a_2, package_b_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_b_2]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_b_1, package_c_1]),
                frozenset([package_b_2, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # and finally with a second c
        packages[package_c_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_1, package_b_1]),
                frozenset([package_a_1, package_b_2]),
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_b_1, package_c_1]),
                frozenset([package_b_1, package_c_2]),
                frozenset([package_b_2, package_c_1]),
                frozenset([package_b_2, package_c_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))


class TestInterestingConcurrentInstallsPublishedPackages(TestCase):
    """Tests for the interesting_concurrent_installs feature.

    a Missing conflict finder can report on 'interesting' possible
    concurrent installs. 
    With a published set it should only classify interesting possible
    concurrent installs as those involving one of the currently published
    packages in it.
    """
        
    def setUp(self):
        """All tests here need a finder."""
        super(TestInterestingConcurrentInstallsPublishedPackages, self).setUp()
        # all these tests need a finder.
        self.finder = MissingConflictFinder()
        # set up published packages for testing with
        self.published_packages = set()
        self.published_packages.add(('a', '1-2', 'i386'))
        self.finder.set_published_packages(self.published_packages)
        self.null_replace_conflicts = {
            'replaces':'',
            'conflicts':'',
            }

    def test_calculate_interesting_possible_concurrent_installs(self):
        finder = MissingConflictFinder()
        # nothing in, nothing out.
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs({}))

    def test_same_name_is_not_concurrently_installable(self):
        # the same package always counts as not concurrently installable.
        packages = {
            ('atomix', '1.2.1-1', 'all'):{},
            ('atomix', '1.2.1-1', 'i386'):{},
            ('atomix', '1:1.2.1-1', 'sparc'):{},
            }
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))

    def test_two_package_names_no_conflict_replace(self):
        # when two separate packages on the same arch
        # do not conflict or replace,
        # they are concurrently installable, and we get back only the 
        # combinations including a published package.
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'i386')
        package_b_2 = ('b', '2-2', 'i386')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))

    def test_two_package_names_one_all_no_conflict_replace(self):
        # when two separate packages on the same arch
        # do not conflict or replace,
        # they are concurrently installable. 'all' counts as the
        # same arch for the purposes of this test, so its nearly a dup
        # of the test_two_package_names_no_conflict_replace test.
        # but there isn't enough actual duplicate intent to 
        # factor out yet usefully.
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'all')
        package_b_2 = ('b', '2-2', 'all')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))

    def test_three_package_names_no_conflict_replace(self):
        # when three separate packages on the same arch
        # do not conflict or replace,
        # they are concurrently installable
        # but we dont care that its a group, we want to know about
        # the individual pairs that can occur.
        # as with two-package scenarios, only combinations that
        # are in conjunction with a currently published package should
        # be reported here.
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'i386')
        package_b_2 = ('b', '2-2', 'i386')
        package_c_1 = ('c', '2-1', 'i386')
        package_c_2 = ('c', '2-2', 'i386')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            package_c_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a present as well.
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_a_2, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # and finally with a second c
        packages[package_c_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_a_2, package_c_1]),
                frozenset([package_a_2, package_c_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))

    def test_three_package_names_one_arch_all_no_conflict_replace(self):
        # when three separate packages 
        # do not explicitly conflict or replace,
        # they are concurrently installable, even if one is on arch: all
        # but we dont care that its a group, we want to know about
        # the individual permutations that can occur.
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'all')
        package_b_2 = ('b', '2-2', 'all')
        package_c_1 = ('c', '2-1', 'i386')
        package_c_2 = ('c', '2-2', 'i386')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            package_c_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a present as well.
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_a_2, package_c_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # and finally with a second c
        packages[package_c_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                frozenset([package_a_2, package_c_1]),
                frozenset([package_a_2, package_c_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))

    def test_three_package_names_two_arch_one_all_no_conflict_replace(self):
        # when three separate packages 
        # do not explicitly conflict or replace, and one is arch X, one is
        # arch Y, one is arch: all, the X and all, and Y and all are
        # concurrently installable.
        # they are concurrently installable, even if one is on arch: all
        # but we dont care that its a group, we want to know about
        # the individual errors that can occur
        package_a_1 = ('a', '1-1', 'i386')
        package_a_2 = ('a', '1-2', 'i386')
        package_b_1 = ('b', '2-1', 'all')
        package_b_2 = ('b', '2-2', 'all')
        package_c_1 = ('c', '2-1', 'sparc')
        package_c_2 = ('c', '2-2', 'sparc')
        packages = {
            package_a_1:self.null_replace_conflicts,
            package_b_1:self.null_replace_conflicts,
            package_c_1:self.null_replace_conflicts,
            }
        self.assertEqual(
            set(),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of a present as well.
        packages[package_a_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # try with another version of b
        packages[package_b_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))
        # and finally with a second c
        packages[package_c_2] = self.null_replace_conflicts
        self.assertEqual(
            set([
                frozenset([package_a_2, package_b_1]),
                frozenset([package_a_2, package_b_2]),
                ]),
            self.finder.interesting_concurrent_installs(packages))


class TestGetRelationShip(TestCase):
    """Test that the finder get_relationship attribute is configurable.

    We use this configuration in later tests.
    """

    def test_default_get_relationship(self):
        # the default get_relationship should work
        finder = MissingConflictFinder()
        self.assertEqual(finder.get_relationship, DebRelationship)

    def test_set_get_relationship(self):
        # we should be able to override get_relationship for testing.
        finder = MissingConflictFinder()
        finder.set_get_relationship(list)
        self.assertEqual(finder.get_relationship, list)
        self.assertEqual([], finder.get_relationship())


class MockRelationship(object):
    """A mock relationship which record its actions for testing."""

    def __init__(self, results):
        """prime the relationship with the results to give."""
        self.results = results
        self._matches = []

    def matches(self, package, version, provides=''):
        self._matches.append((package, version, provides))
        return self.results.pop(0);


class TestConcurrentPermitted(TestCase):
    """Tests for the concurrent_permitted checks."""

    def test_concurrent_permitted(self):
        # the concurrency permitted routine should create and evaluate the
        # conflicts clauses, with boolean short circuiting.
        # -it is equivalent to return not (a.matches(b) || b.matches(a))
        finder = MissingConflictFinder()
        package_a = {
            'name':'a',
            'version':'1-1',
            'arch':'all',
            'conflicts':'c',
            'replaces':'e',
            'provides':'g',
            }
        package_b = {
            'name':'b',
            'version':'2-2',
            'arch':'all',
            'conflicts':'d',
            'replaces':'f',
            'provides':'h',
            }
        # prime the mock relationship factory
        # returns false

        relations_made = []
        relationship = MockRelationship([False, False])
        def get_relationship(relation):
            relations_made.append(relation)
            return relationship

        finder.set_get_relationship(get_relationship)
        # as this returns False, False, we expect
        # a construct of 'c', match against 'b', '2-2' -> False
        # a construct of 'd', match against 'a', '1-1' -> False
        self.assertTrue(finder.concurrent_permitted(package_a, package_b))
        self.assertEqual(['c', 'd'], relations_made)
        self.assertEqual(
            [('b', '2-2', 'h'), ('a', '1-1', 'g')],
            relationship._matches)
        self.assertEqual([], relationship.results)
        
        #now test if the second one matches, we get False - concurrency is
        # not permitted
        relationship = MockRelationship([False, True])
        relations_made = []
        # as this returns False, True, we expect
        # a construct of 'c', match against 'b', '2-2' -> False
        # a construct of 'd', match against 'a', '1-1' -> True
        self.assertFalse(finder.concurrent_permitted(package_a, package_b))
        self.assertEqual(['c', 'd'], relations_made)
        self.assertEqual(
            [('b', '2-2', 'h'), ('a', '1-1', 'g')],
            relationship._matches)
        self.assertEqual([], relationship.results)
        
        # finally, if the first one tested returned true, the evaluation
        # should short-circuit and we want False
        relationship = MockRelationship([True, False])
        relations_made = []
        # as this returns False, True, we expect
        # a construct of 'c', match against 'b', '2-2' -> True
        self.assertFalse(finder.concurrent_permitted(package_a, package_b))
        self.assertEqual(['c'], relations_made)
        self.assertEqual(
            [('b', '2-2', 'h')],
            relationship._matches)
        self.assertEqual([False], relationship.results)


class TestConcurrentSafe(TestCase):
    """Tests for the concurrent_safe checks."""

    def test_concurrent_safe(self):
        # the concurrency safe routine should create and evaluate the
        # replaces clauses, with boolean short circuiting.
        # - it is equivalent to return (a.matches(b) || b.matches(a)) on the
        # replaces clause
        finder = MissingConflictFinder()
        package_a = {
            'name':'a',
            'version':'1-1',
            'arch':'all',
            'conflicts':'c',
            'replaces':'e',
            'provides':'g',
            }
        package_b = {
            'name':'b',
            'version':'2-2',
            'arch':'all',
            'conflicts':'d',
            'replaces':'f',
            'provides':'h',
            }
        # prime the mock relationship factory
        relations_made = []
        relationship = MockRelationship([False, False])
        def get_relationship(relation):
            relations_made.append(relation)
            return relationship

        finder.set_get_relationship(get_relationship)
        # as this returns False, False, we expect
        # a construct of 'e', match against 'b', '2-2' -> False
        # a construct of 'f', match against 'a', '1-1' -> False
        self.assertFalse(finder.concurrent_safe(package_a, package_b))
        self.assertEqual(['e', 'f'], relations_made)
        self.assertEqual(
            [('b', '2-2', ''), ('a', '1-1', '')],
            relationship._matches)
        self.assertEqual([], relationship.results)
        
        #now test if the second one matches, we get False - concurrency is
        # not safe
        relationship = MockRelationship([False, True])
        relations_made = []
        # as this returns False, True, we expect
        # a construct of 'e', match against 'b', '2-2' -> False
        # a construct of 'f', match against 'a', '1-1' -> True
        self.assertTrue(finder.concurrent_safe(package_a, package_b))
        self.assertEqual(['e', 'f'], relations_made)
        self.assertEqual(
            [('b', '2-2', ''), ('a', '1-1', '')],
            relationship._matches)
        self.assertEqual([], relationship.results)
        
        # finally, if the first one tested returned true, the evaluation
        # should short-circuit and we want False
        relationship = MockRelationship([True, False])
        relations_made = []
        # as this returns True, False, we expect
        # a construct of 'e', match against 'b', '2-2' -> False
        # a construct of 'f', match against 'a', '1-1' -> False
        self.assertTrue(finder.concurrent_safe(package_a, package_b))
        self.assertEqual(['e'], relations_made)
        self.assertEqual(
            [('b', '2-2', '')],
            relationship._matches)
        self.assertEqual([False], relationship.results)


# are these problems list ?
# I have a replaces: question. If I have a package foo_1-1_i386, then decide to
# move the arch generic stuff out into foo-doc_1-2_all, I use replaces: foo to
# tell dpkg to transfer ownership of the files.. and upload foo_1-2_i386 
# without the doc files. what happens if a user installs foo-doc_1-2_all and 
# then tries to install foo_1-1_i386 ?
# answer: dpkg will note that foo-doc has a replaces and will silently 
#         do the right thing.

# TODO list
# 
# test two package names, one arch, one all, the all replaces the arch -
#      this allow concurrent installs, but is not an error.
#      perhaps we need a second pass to determine if an error exists.
#      i.e. A is arch spec, B is all replaces: A. if B is installed, and I try
#      to install A before the split, does dpkg allow it or is it an error?
#      test with a few combinations of versions, 2 bin arches present etc.
#      should retun a set with the two present in.
# test two package names, same arch, one version of A conflicts with B,
#      but an older version of A does not conflict with B. should return one
#      problematic set - A-that-does-not-conflict-with-B, and B
# test two package names, same arch, conflicts and replaces, not concurrent
# test versioned conflicts clauses.
# test three package names, two one different arches, one on same. -
#      returns two sets, both with the odd one out package, and one with each
#      of the two local packages.
# test three package names, one arch all, two on different archves - 
#      same as 3 packages, 2 arches present.
# test three package names, 3 arches, not concurrent
# test three package names, one arch all, two on different arches,
#      the arch all replaces the arch specific packages,
#      want back two conflict sets, one for all+archA, and all+archB
# test three package names, replaces + conflicts - not concurrent
# test three package names, conflicts - not concurrent.
# three package names, two conflict mutually, third conflicts with one 
#       unidirectionally. should return one set containins the one that the
#       third does not not conflict with, and the third should be returned 
#       as one set. This will show up use of cross product rather than 
#       combinations
# three names, 2 in all, 1 specific - should get crosses within the all set.
# 2 packages names, but changes arch in one version to all.
#
# test two package names, same arch, one version of A conflicts with B,
#      but an older version of A does not conflict with B. should return one
#      problematic set - A-that-does-not-conflict-with-B, and B
# 
# virtual packages? foo provides mta, bar provides mta, and both conflict mta
#                   so we need to consider provides in the conflicts 
#                   resolution.
#                   this may require changing the representation for 
#                   evaluate_conflicts
#
# three packages - two versions of A, conflict on the same file, one version
# of B that conflicts with only one version of A, should not find B and the 
# other A as a conflict. we do this by returning a dictionary from the
# add_files method (files : package lists).
# TODO: also start to only considering the 
# conflicts with the package
# we are adding - which will reduce the checks from a N*M to a M size and 
# is thus good anyway. If we do that, we should have a means to generate
# good stats from the system anyway - perhaps by tricky reporting, or 
# perhaps by implementing both fixes - one fix to ensure that get back
# only fully-conflicting sets from add_files, and one to only consider 
# conflicts with the new package as an optional tweak to interesting
# concurrent_installs.
