# Copyright (C) 2006 Canonical Limited.
#
# Licence to be determined
#

"""Tests for the main() driver."""

from cStringIO import StringIO
import gc
import os.path
import shutil
import sys
import tempfile
from unittest import TestCase

from bzrlib import progress
from storm.exceptions import IntegrityError
import storm.locals

import findconflicts
from findconflicts import (
    ArchiveDatabase,
    ComponentGroupingHandler,
    Filepath,
    MissingConflictFinder,
    PackageArchitecture,
    PackageDataSource,
    PackageFile,
    PackageName,
    PackageVersion,
    )
from findconflicts.main import Main
from findconflicts.tests.test_persistence import FakeConflictHandler


class StubDataSource(PackageDataSource):
    """An implementation of PackageDataSource sufficient to test main()."""

    def __init__(self):
        PackageDataSource.__init__(self)
        self.calls = []
        self._url_info = {
            'foo-v1-deb-url':{
                'arch':'i386',
                'controltext':'controltext',
                'name':'foo',
                'version':'1.1',
                'files':['/etc/foo'],
                },
            'foo-v2-deb-url':{
                'arch':'i386',
                'controltext':'controltext',
                'name':'foo',
                'version':'1.2',
                'files':['/etc/foo'],
                },
            'quux-v1-deb-url':{
                'arch':'sparc',
                'controltext':'controltext',
                'name':'quux',
                'version':'1.1',
                'files':['/etc/quux', '/etc/foo'],
                },
            }

    def get_all_packages(self):
        self.calls.append(('get_all_packages', ))
        return {
            ('foo', '1.2', 'i386'):'foo-v2-deb-url',
            ('foo', '1.1', 'i386'):'foo-v1-deb-url',
            ('quux', '1.1', 'sparc'):'quux-v1-deb-url',
            }, {
            ('foo', '1.2', 'i386'):['fiesty/main'],
            ('foo', '1.1', 'i386'):['dapper/main'],
            ('quux', '1.1', 'sparc'):['feisty/main'],
            }

    def infoFromPackageUrl(self, url):
        self.calls.append(('infoFromPackageUrl', url))
        return self._url_info[url]


class StubArchiveDatabase(ArchiveDatabase):
    """A pretend ArchiveDatabase for testing."""

    @staticmethod
    def open_or_create():
        result = ArchiveDatabase.initialize('sqlite:')
        result.isstubdb = True
        return result


class TrackingPackageVersion(PackageVersion):

    @classmethod
    def Analyse_ids(klass, package_ids, store, finder, pkg_components, log, handler):
        klass.calls.append(('Analyse_ids', package_ids, store, finder, pkg_components, log, handler))
        return PackageVersion.Analyse_ids(package_ids, store, finder, pkg_components, log, handler)

    @classmethod
    def Find_new_versions(klass, package_tuples, store):
        klass.calls.append(('Find_new_versions', package_tuples, store))
        return PackageVersion.Find_new_versions(package_tuples, store)

    @classmethod
    def InsertFromInfo(klass, info, store):
        klass.calls.append(('InsertFromInfo', info, store))
        return PackageVersion.InsertFromInfo(info, store)

    @classmethod
    def To_analyze(klass, store, published_packages):
        klass.calls.append(('To_analyze', store, published_packages))
        return PackageVersion.To_analyze(store, published_packages)


class TestMain(TestCase):

    def test_default_construct(self):
        """Check the default helper objects for main are the expected types."""
        # testing the store is the right one is harder, as we dont want to 
        # chroot around, and construction can creat a file on disk. 
        # so we monkey patch the module for the test duration
        # XXX: Perhaps we should not create the store object
        #      until run(), because that way we can accept parameters to
        #      control the store URI.
        old_db = findconflicts.main.ArchiveDatabase
        findconflicts.main.ArchiveDatabase = StubArchiveDatabase
        try:
            main = Main()
        finally:
            findconflicts.main.ArchiveDatabase = old_db
        self.assertTrue(isinstance(main.source, PackageDataSource))
        self.assertTrue(isinstance(main.finder, MissingConflictFinder))
        self.assertTrue(main.archive_db is StubArchiveDatabase)
        self.assertTrue(main.store.isstubdb)
        self.assertTrue(main.PackageVersion is PackageVersion)
        self.assertTrue(isinstance(main.progress, progress._BaseProgressBar))
        self.assertTrue(sys.stdout is main.log)
        self.assertTrue(isinstance(main.handler, ComponentGroupingHandler))

    def test_replace_archivedatabase_construct(self):
        """We should be able to provide a replacement db layer on the fly."""
        main = Main(archive_db=StubArchiveDatabase)
        self.assertTrue(main.store.isstubdb)

    def test_actions(self):
        """Test main actions:

        first update the package database.
        second after inserting new packages do an analysis run.
        """
        # todo test progress bar use too.
        main = Main(archive_db=StubArchiveDatabase)
        main.handler = FakeConflictHandler()
        main.log = StringIO()
        main.source = StubDataSource()
        # foo-v1-deb is already in the system for this test.
        PackageVersion.InsertFromInfo(main.source._url_info['foo-v1-deb-url'],
            main.store)
        main.PackageVersion = TrackingPackageVersion
        TrackingPackageVersion.calls = []
        main.run(['findconflicts'])
        # one call to get all packages, then one call for each new 
        # package.
        self.assertEqual([('get_all_packages',),
            ('infoFromPackageUrl', 'foo-v2-deb-url'),
            ('infoFromPackageUrl', 'quux-v1-deb-url')],
            main.source.calls)
        self.assertEqual(('Find_new_versions',
            {('foo', '1.1', 'i386'): 'foo-v1-deb-url',
             ('quux', '1.1', 'sparc'): 'quux-v1-deb-url',
             ('foo', '1.2', 'i386'): 'foo-v2-deb-url'},
            TrackingPackageVersion.calls[0][2]), # cant compare the store
            TrackingPackageVersion.calls[0])
        self.assertTrue(TrackingPackageVersion.calls[0][2].isstubdb)
        # now we want to check that it added each new package via 
        # infoFromPackageUrl + InsertFromInfo
        self.assertEqual(('InsertFromInfo',
            main.source._url_info['foo-v2-deb-url'],
            TrackingPackageVersion.calls[1][2]), # cant compare the store
            TrackingPackageVersion.calls[1])
        self.assertTrue(TrackingPackageVersion.calls[1][2].isstubdb)
        self.assertEqual(('InsertFromInfo',
            main.source._url_info['quux-v1-deb-url'],
            TrackingPackageVersion.calls[2][2]), # cant compare the store
            TrackingPackageVersion.calls[2])
        self.assertTrue(TrackingPackageVersion.calls[2][2].isstubdb)
        self.assertEqual(('To_analyze',
            TrackingPackageVersion.calls[3][1], # cant compare the store
             {('foo', '1.1', 'i386'): 'foo-v1-deb-url',
              ('quux', '1.1', 'sparc'): 'quux-v1-deb-url',
              ('foo', '1.2', 'i386'): 'foo-v2-deb-url'},),
            TrackingPackageVersion.calls[3])
        self.assertTrue(TrackingPackageVersion.calls[3][1].isstubdb)
        self.assertEqual(('Analyse_ids',
            [1, 2, 3], # returned from analyze ids.
            TrackingPackageVersion.calls[4][2], # cant compare the store
            main.finder,
            {('foo', '1.1', 'i386'): ['dapper/main'],
             ('quux', '1.1', 'sparc'): ['feisty/main'],
             ('foo', '1.2', 'i386'): ['fiesty/main']},
            main.log,
            main.handler),
            TrackingPackageVersion.calls[4])
        self.assertTrue(TrackingPackageVersion.calls[4][2].isstubdb)
        # there should only have been three calls to TrackingPackageVersion.
        self.assertEqual(5, len(TrackingPackageVersion.calls))
        TrackingPackageVersion.calls = []

