# Copyright (C) 2006 Canonical Limited.
#
# Licence to be determined
#

"""Tests for parseing of preinst files."""

from unittest import TestCase

from findconflicts import (
    ParseError,
    PreinstParser,
    )


class TestPreinstParser(TestCase):

    def setUp(self):
        TestCase.setUp(self)
        self.parser = PreinstParser()

    def test_extract_not_shell(self):
        self.assertRaises(ParseError, self.parser.extract_diversions,
            "#!/usr/bin/python\n")

    def test_recognise_shell_variants(self):
        self.assertEqual([], self.parser.extract_diversions("#!/bin/sh\n"))
        self.assertEqual([], self.parser.extract_diversions("#! /bin/sh\n"))
        self.assertEqual([], self.parser.extract_diversions("#!/bin/bash\n"))
        self.assertEqual([], self.parser.extract_diversions("#! /bin/bash\n"))

    def test_ignores_list(self):
        self.assertEqual([], self.parser.extract_diversions(
            "#! /bin/sh\n"
            "divs=`dpkg-divert --list slocate`\n"
            ))

    def test_ignore_removes(self):
        self.assertEqual([], self.parser.extract_diversions(
            "#! /bin/sh\n"
            "$SL_DEBUG dpkg-divert --package slocate --remove --rename \\\n"
            "  --divert /etc/cron.daily.find.$suffix \n"
            ))

    def test_handles_multiline_strings(self):
        self.assertEqual(['foo'], self.parser.extract_diversions(
            "#!/bin/sh\n"
            "FOO=\"bar\n"
            "      gam\"\n"
            " dpkg-divert --package foo --add --rename \\\n"
            "    --divert foo.notreal foo\n"
            ))

    def test_handles_trap_escaping(self):
        self.assertEqual(['foo'], self.parser.extract_diversions(
            "#!/bin/sh\n"
            "trap \"message;\\\n"
            "       message \\\"Received signal.  Aborting script $0.\\\";\\\n"
            "       message;\\\n"
            "       exit 1\" 1 2 3 15\n"
            " dpkg-divert --package foo --add --rename \\\n"
            "    --divert foo.notreal foo\n"
            ))

    def test_ignores_commented_diversions(self):
        self.assertEqual([],
            self.parser.extract_diversions("""#!/bin/sh
    # dpkg-divert --package foo --add --rename \\
    #             --divert /usr/bin/other.real /usr/bin/other
"""))
        # second variation because currently \ is not processed
        # correctly on commented out lines, and this lets the 
        # regex engine still match.
        self.assertEqual([],
            self.parser.extract_diversions("""#!/bin/sh
    # dpkg-divert --package foo --add --rename \\
                 --divert /usr/bin/other.real /usr/bin/other
"""))

    def test_find_slocate_diversions(self):
        self.assertEqual(
            ['/usr/bin/locate',
             '/usr/bin/updatedb',
             '/usr/share/man/man1/locate.1.gz',
             '/usr/share/man/man1/updatedb.1.gz',
             '/etc/cron.daily/find',
            ],
            self.parser.extract_diversions("""#!/bin/sh
#if [ install = "$1" ]; then
	$SL_DEBUG dpkg-divert --package slocate --add --rename \\
		--divert /usr/bin/locate.$suffix \\
			/usr/bin/locate
	$SL_DEBUG dpkg-divert --package slocate --add --rename \\
		--divert /usr/bin/updatedb.$suffix \\
			/usr/bin/updatedb
	$SL_DEBUG dpkg-divert --package slocate --add --rename \\
		--divert /usr/share/man/man1/locate.$suffix.1.gz \\
			/usr/share/man/man1/locate.1.gz
	$SL_DEBUG dpkg-divert --package slocate --add --rename \\
		--divert /usr/share/man/man1/updatedb.$suffix.1.gz \\
			/usr/share/man/man1/updatedb.1.gz
	# The cron script must be moved out of cron.daily so that it is
	# no longer run.
	$SL_DEBUG dpkg-divert --package slocate --add --rename \\
		--divert /etc/cron.daily/find.$suffix \\
			/etc/cron.daily/find
#fi
"""))

    def test_find_gimp_print_diversions(self):
        self.assertEqual(
            ['/usr/lib/gimp/2.0/plug-ins/print',
            ],
            self.parser.extract_diversions("""#!/bin/sh
           dpkg-divert --package gimp-print --add --rename \\
	      --divert /usr/lib/gimp/2.0/print.orig \\
	      /usr/lib/gimp/2.0/plug-ins/print
"""))

    def test_find_bitchx_ssl_diversions(self):
        self.assertEqual(
            ['/usr/bin/BitchX',
             '/usr/share/man/man1/BitchX.1.gz',
            ],
            self.parser.extract_diversions("""#!/bin/sh
	dpkg-divert --package bitchx-ssl --divert /usr/bin/BitchX.nossl \\
		--rename --add /usr/bin/BitchX
	dpkg-divert --package bitchx-ssl --divert /usr/share/man/man1/BitchX.nossl.1.gz \\
		--rename --add /usr/share/man/man1/BitchX.1.gz
"""))
