# Copyright (C) 2006 Canonical Limited.
#
# Licence to be determined
#

"""Tests for persistent data storage."""

from cStringIO import StringIO
import gc
import os.path
import shutil
import tempfile
from unittest import TestCase

from storm.exceptions import IntegrityError
import storm.locals

from findconflicts import (
    ArchiveDatabase,
    Filepath,
    MissingConflictFinder,
    PackageArchitecture,
    PackageFile,
    PackageName,
    PackageVersion,
    )


class FakeConflictHandler(object):

    def __init__(self):
        """Create a FakeConflictHandler."""
        self._calls = []

    def conflict(self, pkg_a, pkg_b):
        self._calls.append(("conflict", pkg_a, pkg_b))

    def start(self):
        self._calls.append(("start", ))

    def finish(self):
        self._calls.append(("finish", ))


class FakeMissingConflictFinder(object):

    def __init__(self, results):
        """Create a FakeMissingConflictFinder.

        :param results: The results to be returned from calls.
        """
        self._calls = []
        self._results = results

    def concurrent_permitted(self, package_left, package_right):
        self._calls.append(('concurrent_permitted', package_left, package_right))
        return self._results.pop(0)

    def concurrent_safe(self, package_left, package_right):
        self._calls.append(('concurrent_safe', package_left, package_right))
        return self._results.pop(0)


class FakePackageVersion(object):
    """A fake packageversion for testing top level logic"""

    def __init__(self, results):
        self._calls = []
        self._results = results

    def analyse(self, finder):
        self._calls.append(('analyse', finder))
        return self._results.pop(0)

    def format_conflict(self, other, pkg_components):
        self._calls.append(('format_conflict', other, pkg_components))
        return self._results.pop(0)


class FakeStore(object):
    """A fake store to facilitate top level tests."""

    def __init__(self, results):
        """Create a FakeStore.

        :param results: The results to be returned from calls.
        """
        self._calls = []
        self._results = results
        self._connection = self

    def commit(self):
        self._calls.append(('commit', ))

    def find(self, klass, id):
        self._calls.append(('find', klass, id))
        return self._results.pop(0)

    def flush(self):
        self._calls.append(('flush', ))


class FakeSelect(object):
    """A stub 'select' object for FakeStore results."""

    def __init__(self, result):
        self.result = result

    def one(self):
        return self.result


class TestStore(TestCase):

    def test_initialize_disk(self):
        tempdir = tempfile.mkdtemp()
        try:
            # this checks that the initialize routine, given a uri, works.
            store = ArchiveDatabase.initialize('sqlite:' + tempdir+  '/test.db')
            self.assertTrue(isinstance(store, storm.locals.Store))
            # we dont check for database tables etc here, we will test those
            # with an in memory database on a per-table basis.
            self.assertTrue(os.path.exists(tempdir+  '/test.db'))
        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

    def test_open_or_create_missing(self):
        """When no db exists, the UI will want to 'just work'."""
        tempdir = tempfile.mkdtemp()
        try:
            curdir = os.getcwdu()
            os.chdir(tempdir)
            try:
                # this checks that open_or_create creates a findconflicts.db
                # when none exists.
                store = ArchiveDatabase.open_or_create()
                self.assertTrue(os.path.exists(tempdir + '/findconflicts.db'))
                # and we can smoke check it has the expected contents by doing
                # a simple query
                self.assertEqual([], list(store.find(PackageName)))
            finally:
                os.chdir(curdir)
        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

    def test_open_or_create_existing(self):
        """When not initializing, the UI will want to 'just work'."""
        tempdir = tempfile.mkdtemp()
        try:
            curdir = os.getcwdu()
            os.chdir(tempdir)
            try:
                # this checks that open_or_create opens a findconflicts.db
                # when none exists without removing its data
                # add data
                store = ArchiveDatabase.open_or_create()
                PackageName.Ensure('demo', store)
                store.commit()
                del store
                gc.collect()
                # open or create anew
                store = ArchiveDatabase.open_or_create()
                self.assertTrue(os.path.exists(tempdir + '/findconflicts.db'))
                # and we can smoke check it has the expected contents by doing
                # a simple query
                self.assertEqual([PackageName('demo')],
                    list(store.find(PackageName)))
            finally:
                os.chdir(curdir)
        finally:
            shutil.rmtree(tempdir, ignore_errors=True)


class TestArchitectures(TestCase):

    def test___eq__(self):
        self.assertNotEqual(None, PackageArchitecture('i386'))
        self.assertNotEqual(PackageArchitecture('i386'),
            PackageArchitecture('all'))
        self.assertEqual(PackageArchitecture('all'),
            PackageArchitecture('all'))

    def test_initialize(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # a new db should have no known architectures
        self.assertEqual([], list(store.find(PackageArchitecture)))

    def test_fields(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # test we can insert objects with the expected fields,
        # and get them back.
        i386 = PackageArchitecture('i386')
        store.add(i386)
        # we should now have it safely in the database. committing 
        # and doing a garbage collect should let us get at it 
        # uncached.
        store.commit()
        del i386
        gc.collect()
        self.assertEqual([PackageArchitecture('i386')],
            list(store.find(PackageArchitecture)))
        # we should also be unable to add duplicates.
        newi386 = PackageArchitecture('i386')
        store.add(newi386)
        self.assertRaises(IntegrityError, store.commit)
        # having made it fail, rollback.
        store.rollback()
        # and check its still accessible.
        self.assertEqual([PackageArchitecture('i386')],
            list(store.find(PackageArchitecture)))
        # we should be able to add another arch at this point.

    def test_Ensure(self):
        """Ensure should provide an easy api for working with Architectures."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # create a PackageArchitecture in the store.
        _usr_bin_mail = PackageArchitecture.Ensure('all', store)
        store.commit()
        self.assertEqual([PackageArchitecture('all')],
            list(store.find(PackageArchitecture)))
        # and reuse an existing one if possible.
        a = PackageArchitecture.Ensure('i386', store)
        b = PackageArchitecture.Ensure('i386', store)
        self.assertTrue(a is b)


class TestPackageNames(TestCase):

    def test___eq__(self):
        self.assertNotEqual(None, PackageName('foo'))
        self.assertNotEqual(PackageName('bar'),
            PackageName('foo'))
        self.assertEqual(PackageName('foo'),
            PackageName('foo'))

    def test_initialize(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # a new db should have no known package names
        self.assertEqual([], list(store.find(PackageName)))

    def test_fields(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # test we can insert objects with the expected fields,
        # and get them back.
        atomix = PackageName('atomix')
        store.add(atomix)
        # we should now have it safely in the database. committing 
        # and doing a garbage collect should let us get at it 
        # uncached.
        del atomix
        gc.collect()
        store.commit()
        self.assertEqual([PackageName('atomix')],
            list(store.find(PackageName)))
        # we should also be unable to add duplicates.
        newatomix = PackageName('atomix')
        store.add(newatomix)
        self.assertRaises(IntegrityError, store.commit)
        # having made it fail, rollback.
        store.rollback()
        self.assertEqual([PackageName('atomix')],
            list(store.find(PackageName)))

    def test_versions(self):
        """PackageName.versions gives the versions of a PackageName."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        all = PackageArchitecture.Ensure('all', store)
        atomix = PackageName.Ensure('atomix', store)
        self.assertEqual(0, atomix.versions.count())
        a_version = PackageVersion(atomix, '3', all, '')
        store.add(a_version)
        store.flush()
        self.assertEqual(set([a_version]), set(atomix.versions))
        # check how we should write 'find' clauses.
        # add another architecture.
        i386 = PackageArchitecture.Ensure('i386', store)
        # and create a version for i386, not linked to any
        # store objects, to avoid automatic-adding.
        b_version = PackageVersion(PackageName('atomix'),
            '3',
            PackageArchitecture('i386'),
            '')
        self.assertTrue(a_version in atomix.versions)
        self.assertFalse(b_version in atomix.versions)
        # smaller sets are good for probing: limit the versions examined.
        self.assertTrue(a_version in atomix.versions.find(version='3'))
        self.assertFalse(b_version in atomix.versions.find(version='3'))
        # negative check for the version filtering - should not return
        # the '3' versions.
        self.assertFalse(a_version in atomix.versions.find(version='2'))

    def test_Ensure(self):
        """Ensure should provide an easy api for working with PackageNames."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # create a PackageName in the store.
        _usr_bin_mail = PackageName.Ensure('atomix', store)
        store.commit()
        self.assertEqual([PackageName('atomix')],
            list(store.find(PackageName)))
        # and reuse an existing one if possible.
        a = PackageName.Ensure('915resolution', store)
        b = PackageName.Ensure('915resolution', store)
        self.assertTrue(a is b)


class TestPackageVersions(TestCase):

    def test___eq__(self):
        self.assertNotEqual(None,
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''))
        # different name
        self.assertNotEqual(
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''),
            PackageVersion(PackageName('bar'), '3', PackageArchitecture('all'),
                ''))
        # different arch
        self.assertNotEqual(
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''),
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('i386'),
                ''))
        # different version
        self.assertNotEqual(
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''),
            PackageVersion(PackageName('foo'), '2', PackageArchitecture('all'),
                ''))
        # different controltext
        self.assertNotEqual(
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                'abc'),
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''))
        # all the same
        self.assertEqual(
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''),
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''))
        # different analyzed_level
        self.assertNotEqual(
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                '', analyzed_level=1),
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''))
        # different preinst hash
        self.assertNotEqual(
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                ''),
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                '', preinst_hash="foo"))
        self.assertNotEqual(
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                '', preinst_hash="bar"),
            PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
                '', preinst_hash="foo"))

    def test_analyse(self):
        # analyse is the top level driver. 
        # analyse should grab the packages to check against and 
        # then use conflicts_against_package for each.
        # we want analyse to return the remaining problematic ones
        # and to set self.analyzed_level as appropriate. 
        # there are two cases: either the set of conflicts is empty or not.
        # We test with a PackageVersion object which we have tweaked its 
        # find_unsafely_colocated_packages and conflicts_against_package 
        # methods to be stubs we can provide canned results from to test
        # these cases easily.
        store = ArchiveDatabase.initialize('sqlite:')
        # add foo-1.2-all and quux-1.3-i386 - our two packages to use.
        all = PackageArchitecture.Ensure('all', store)
        i386 = PackageArchitecture.Ensure('i386', store)
        foo = PackageName.Ensure('foo', store)
        quux = PackageName.Ensure('quux', store)
        foo_version = PackageVersion(foo, '1.2', all, '')
        quux_version = PackageVersion(quux, '1.3', i386, '')
        foo_version.controltext = (
            "Version: foo\n"
            "Conflicts: quux\n")
        quux_version.controltext = (
            "Version: quux\n"
            "Conflicts: foo\n")
        store.add(foo_version)
        store.add(quux_version)
        store.flush()
        finder = MissingConflictFinder()
        # setup the instrumented foo_version methods.
        def find_packages():
            return [quux_version]
        foo_version.find_unsafely_colocated_packages = find_packages
        def conflicts_against_package(packageversion, finder):
            return False
        foo_version.conflicts_against_package = conflicts_against_package
        # when no conflicts are found, the package being analysed has its
        # analyzed_level set to 1.
        foo_version.analyzed_level = 0
        self.assertEqual(set([]), set(foo_version.analyse(finder)))
        self.assertEqual(1, foo_version.analyzed_level)
        # setup new instrumented versions for the case where there are
        # conflicts.
        def conflicts_against_package(packageversion, finder):
            return True
        foo_version.conflicts_against_package = conflicts_against_package
        # when conflicts occur, they are returned, and analyzed_level is 
        # untouched.
        foo_version.analyzed_level = 0
        self.assertEqual(set([quux_version]), set(foo_version.analyse(finder)))
        self.assertEqual(0, foo_version.analyzed_level)
        foo_version.analyzed_level = 1
        self.assertEqual(set([quux_version]), set(foo_version.analyse(finder)))
        self.assertEqual(1, foo_version.analyzed_level)

    def test_Analyse_ids_none(self):
        store = FakeStore([])
        handler = FakeConflictHandler()
        log = StringIO()
        # nothing to do - should note there are no conflicts.
        PackageVersion.Analyse_ids([], store, 'finder', {}, log, handler)
        self.assertEqual("no errors (checked 0 packages).\n",
            log.getvalue())
        self.assertEqual([], store._calls)
        self.assertEqual([('start',), ('finish',)], handler._calls)

    def test_Analyse_ids_noconflicts(self):
        # two non conflicting package - should perform two store commits.
        packagea = FakePackageVersion([[]])
        packageb = FakePackageVersion([[]])
        store = FakeStore([FakeSelect(packagea), FakeSelect(packageb)])
        handler = FakeConflictHandler()
        log = StringIO()
        PackageVersion.Analyse_ids([1, 2], store, 'finder', {}, log, handler)
        self.assertTrue(log.getvalue().endswith(
            "no errors (checked 2 packages).\n"))
        self.assertEqual([
            ('find', PackageVersion, 1),
            ('flush',),
            ('commit',),
            ('find', PackageVersion, 2),
            ('flush',),
            ('commit',)],
            store._calls)
        self.assertEqual([('analyse', 'finder')], packagea._calls)
        self.assertEqual([('analyse', 'finder')], packageb._calls)
        self.assertEqual([('start',), ('finish',)], handler._calls)

    def test_Analyse_ids_oneconflict(self):
        # one conflicting package - should perform a single error log.
        packagea = FakePackageVersion([['packageb'], 'a and b'])
        store = FakeStore([FakeSelect(packagea)])
        handler = FakeConflictHandler()
        log = StringIO()
        pkg_components = {'a':['distro1/component', 'distro2/component']}
        PackageVersion.Analyse_ids([1], store, 'finder', pkg_components, log, handler)
        self.assertEqual([('find', PackageVersion, 1)], store._calls)
        self.assertEqual(
            [('analyse', 'finder')],
            packagea._calls)
        self.assertEqual([('start',),
            ('conflict', packagea, 'packageb'),
            ('finish',)], handler._calls)

    def test_conflicts_against_package(self):
        # The conflicts_against_package method wraps the use of
        # MissingConflictFinder. We test it by checking its use of the 
        # MissingConflictFinder and its return values.
        store = ArchiveDatabase.initialize('sqlite:')
        # add foo-1.2-all and quux-1.3-i386 - our two packages to use.
        all = PackageArchitecture.Ensure('all', store)
        i386 = PackageArchitecture.Ensure('i386', store)
        foo = PackageName.Ensure('foo', store)
        quux = PackageName.Ensure('quux', store)
        foo_version = PackageVersion(foo, '1.2', all, '')
        quux_version = PackageVersion(quux, '1.3', i386, '')
        foo_version.controltext = (
            "Version: foo\n"
            "Conflicts: quux\n")
        quux_version.controltext = (
            "Version: quux\n"
            "Conflicts: foo\n")
        store.add(foo_version)
        store.add(quux_version)
        store.flush()
        finder = FakeMissingConflictFinder([
        # test one - concurrent allowed, concurrent safe -> returns False
            True, True,
        # test two - concurrent allowed, concurrent unsafe -> returns True
            True, False,
        # test three - concurrent not allowed -> returns False
            False
        ])
        self.assertEqual(False, foo_version.conflicts_against_package(quux_version, finder))
        self.assertEqual(
            [('concurrent_permitted', {'replaces': '', 'name': 'foo',
                'version': '1.2', 'provides': '', 'conflicts': 'quux',
                'arch': 'all'},
              {'replaces': '', 'name': 'quux', 'version': '1.3', 'provides': '',
               'conflicts': 'foo', 'arch': 'i386'}),
             ('concurrent_safe', {'replaces': '', 'name': 'foo',
              'version': '1.2', 'provides': '', 'conflicts': 'quux',
              'arch': 'all'}, {'replaces': '', 'name': 'quux', 'version': '1.3',
              'provides': '', 'conflicts': 'foo', 'arch': 'i386'})
            ],
            finder._calls)
        finder._calls = []
        self.assertEqual(True, foo_version.conflicts_against_package(quux_version, finder))
        self.assertEqual(
            [('concurrent_permitted', {'replaces': '', 'name': 'foo',
                'version': '1.2', 'provides': '', 'conflicts': 'quux',
                'arch': 'all'},
              {'replaces': '', 'name': 'quux', 'version': '1.3', 'provides': '',
               'conflicts': 'foo', 'arch': 'i386'}),
             ('concurrent_safe', {'replaces': '', 'name': 'foo',
              'version': '1.2', 'provides': '', 'conflicts': 'quux',
              'arch': 'all'}, {'replaces': '', 'name': 'quux', 'version': '1.3',
              'provides': '', 'conflicts': 'foo', 'arch': 'i386'})
            ],
            finder._calls)
        finder._calls = []
        self.assertEqual(False, foo_version.conflicts_against_package(quux_version, finder))
        self.assertEqual(
            [('concurrent_permitted', {'replaces': '', 'name': 'foo',
                'version': '1.2', 'provides': '', 'conflicts': 'quux',
                'arch': 'all'},
              {'replaces': '', 'name': 'quux', 'version': '1.3', 'provides': '',
               'conflicts': 'foo', 'arch': 'i386'}),
            ],
            finder._calls)

    def test_initialize(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # a new db should have no known package versions
        self.assertEqual([], list(store.find(PackageVersion)))

    def test_fields(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # test we can insert objects with the expected fields,
        # and get them back.
        atomix_3_all = PackageVersion(
            PackageName('atomix'), '3', PackageArchitecture('all'),
            '')
        store.add(atomix_3_all)
        # we should now have it safely in the database. committing 
        # and doing a garbage collect should let us get at it 
        # uncached.
        store.commit()
        del atomix_3_all
        gc.collect()
        # test the store version by inspection, not equality, because
        # the comparison check triggers a flush, and forces our version
        # to compare with to be added, causing an integrity error.
        store_versions = list(store.find(PackageVersion))
        self.assertNotEqual(store_versions, [None])
        store_versions[0].name
        store_versions[0].arch
        expected_name = store.find(PackageName, name='atomix').one()
        expected_version = '3'
        expected_arch = store.find(PackageArchitecture, arch='all').one()
        
        new_atomix_3_all = PackageVersion(
            expected_name,
            expected_version,
            expected_arch,
            '')
        self.assertEqual([new_atomix_3_all], store_versions)
        # we should be unable to add duplicates.
        store.add(new_atomix_3_all)
        self.assertRaises(storm.exceptions.IntegrityError, store.commit)
        # having made it fail, rollback.
        store.rollback()
        self.assertEqual([new_atomix_3_all],
            list(store.find(PackageVersion)))
        atomix_3_all = store.get(PackageVersion, 1)
        # test un unsupplied hash gives us the sha1sum of a empty string(file)
        self.assertEqual("da39a3ee5e6b4b0d3255bfef95601890afd80709", atomix_3_all.preinst_hash)
        # we should be able to add Filepaths to the package version
        foo = Filepath.Ensure('/foo', store)
        bar = Filepath.Ensure('/bar', store)
        atomix_3_all.files.add(foo)
        atomix_3_all.files.add(bar)
        store.commit()
        # get a clean copy
        del atomix_3_all
        gc.collect()
        atomix_3_all = store.get(PackageVersion, 1)
        self.assertEqual(set([foo, bar]), set(atomix_3_all.files))
        # now test if we can insert with a preinst_hash
        atomix_4_all = PackageVersion(
            PackageName('foo'), '4', PackageArchitecture('i386'),
            '', preinst_hash="deadbeef")
        store.add(atomix_4_all)
        store.commit()
        # get a clean copy
        del atomix_4_all
        gc.collect()
        # 2 -> database row id
        atomix_4_all = store.get(PackageVersion, 2)
        self.assertEqual("deadbeef", atomix_4_all.preinst_hash)

    def test_Find(self):
        """PackageVersion has a method to find a package from a tuple."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # a package tuple not in the store
        packagetuple = ('foo', '1.2', 'all')
        self.assertRaises(KeyError, PackageVersion.Find, packagetuple, store)
        # add one component at a time.
        store.add(PackageArchitecture('all'))
        self.assertRaises(KeyError, PackageVersion.Find, packagetuple, store)
        store.add(PackageName('foo'))
        self.assertRaises(KeyError, PackageVersion.Find, packagetuple, store)
        store.add(PackageVersion(PackageName.Ensure('foo', store),
            '1.2',
            PackageArchitecture.Ensure('all', store),
            'content'))
        self.assertEqual(PackageVersion(PackageName('foo'),
            '1.2',
            PackageArchitecture('all'),
            'content'),
            PackageVersion.Find(packagetuple, store))

    def test_Find_new_versions(self):
        """Find the new versions from a set of version tuples."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        self.assertEqual(set(), PackageVersion.Find_new_versions(set(), store))
        # while nothing is in the store, a larger search is returned in full
        candidates = set([
            ('foo', '1.2', 'all'),
            ('foo', '1.3', 'i386'),
            ('quux', '1.3', 'i386')])
        self.assertEqual(candidates, PackageVersion.Find_new_versions(
            candidates, store))
        # things in the store are removed from the input list
        # add foo-1.2-all and quux-1.3-i386
        all = PackageArchitecture.Ensure('all', store)
        i386 = PackageArchitecture.Ensure('i386', store)
        foo = PackageName.Ensure('foo', store)
        quux = PackageName.Ensure('quux', store)
        store.add(PackageVersion(foo, '1.2', all, ''))
        store.add(PackageVersion(quux, '1.3', i386, ''))
        store.flush()
        expected_result = set([('foo', '1.3', 'i386')])
        self.assertEqual(expected_result, PackageVersion.Find_new_versions(
            candidates, store))

    def test_format_conflict(self):
        """A conflict can be nicely formated for human analysis."""
        store = ArchiveDatabase.initialize('sqlite:')
        # create two packages with overlapping files.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/bin/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            }
        foo_package = PackageVersion.InsertFromInfo(info, store)
        # and a second overlapping package.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/bin/foo',
                '/usr/share/doc/bar/copyright',
                ],
            'name':'bar',
            'version':'0.2',
            'arch':'sparc',
            }
        bar_package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual(
            "foo_0.1-0ubuntu2_sparc[not published], bar_0.2_sparc[a/b, c/d] ['/etc/foo', '/usr/bin/foo']",
            foo_package.format_conflict(bar_package, {'bar_0.2_sparc':['a/b', 'c/d']}))

    def test_InsertFromInfo(self):
        """We can take the data from infoFromPackageUrl and insert it."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # info is a dictionary. See findconflicts.tests.test_getpackagesumamry.
        info = {
            'controltext':'controltext',
            'files':[
                '/etc/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'replaces':'rfoo',
            'conflicts':'cfoo',
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            'provides':'pfoo',
            'diversions':['/usr/bin/foo'],
            }
        package = PackageVersion.InsertFromInfo(info, store)
        # now we should have a packageversion linked to the right files.
        def check_package(package):
            self.assertEqual('foo_0.1-0ubuntu2_sparc', str(package))
            self.assertEqual(0, package.analyzed_level)
            self.assertEqual('controltext', package.controltext)
            self.assertEqual([
                Filepath('/etc/foo'),
                Filepath('/usr/share/doc/foo/copyright')],
                sorted(list(package.files)))
            self.assertEqual(1, package.id)
            self.assertEqual([('/usr/bin/foo', )],
                store.execute(
                    "SELECT path FROM packageversion, diversion, filepath "
                    "WHERE packageversion.id=1 AND "
                    "packageversion.id=diversion.package_id AND "
                    "diversion.file_id=filepath.id").get_all())
            self.assertEqual(set(['/usr/bin/foo']), package.diversions)
        check_package(package)
        # reload from disk
        del package
        gc.collect()
        package = store.find(PackageVersion, id=1).one()
        check_package(package)
        # inserting with no diversions should not blowup.
        info = {
            'controltext':'controltext',
            'files':[
                '/etc/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu3',
            'arch':'sparc',
            }
        package = PackageVersion.InsertFromInfo(info, store)
        # inserting with an empty diversion list should not blowup.
        info = {
            'controltext':'controltext',
            'files':[
                '/etc/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu4',
            'arch':'sparc',
            'diversions':[],
            }
        package = PackageVersion.InsertFromInfo(info, store)
        # if preinst_hash is in info dict, it should be added to the package
        info = {
            'controltext':'controltext',
            'files':[],
            'name':'foo',
            'version':'0.1-0ubuntu5',
            'arch':'sparc',
            'preinst_hash':'foobarbaz'
            }
        package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual('foobarbaz', package.preinst_hash)

    def test_find_unsafely_colocated_packages_sole_package(self):
        """When there is one package in the store, no colocation happens."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # cheap way to insert a package.
        info = {
            'controltext':'controltext',
            'files':[
                '/etc/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            }
        package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual([], package.find_unsafely_colocated_packages())

    def test_find_unsafely_colocated_packages_candidates_not_overlapping(self):
        """With non-colated packages on the same arch, nothing colocates."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # cheap way to insert a package.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            }
        PackageVersion.InsertFromInfo(info, store)
        # cheap way to insert a package.
        info = {
            'controltext':'',
            'files':[
                '/etc/bar',
                '/usr/share/doc/bar/copyright',
                ],
            'name':'bar',
            'version':'0.2',
            'arch':'all',
            }
        bar_package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual([], bar_package.find_unsafely_colocated_packages())

    def test_find_unsafely_colocated_packages_candidates_irrelevant_arch(self):
        """With colated packages on the irrelevant arches, nothing happens."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # cheap way to insert a package.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            }
        PackageVersion.InsertFromInfo(info, store)
        # cheap way to insert a package.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/share/doc/bar/copyright',
                ],
            'name':'bar',
            'version':'0.2',
            'arch':'i386',
            }
        bar_package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual([], bar_package.find_unsafely_colocated_packages())

    def test_find_unsafely_colocated_packages_candidates_relevant_arches(self):
        """With colocated packages on relevant arches, a list is returned."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # cheap way to insert a package.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            }
        foo_package = PackageVersion.InsertFromInfo(info, store)
        # a alternate package for this one, on the same arch.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/share/doc/bar/copyright',
                ],
            'name':'bar',
            'version':'0.2',
            'arch':'sparc',
            }
        bar_package = PackageVersion.InsertFromInfo(info, store)
        # and another alternate, arch all - conflicts with anything.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/share/doc/gam/copyright',
                ],
            'name':'gam',
            'version':'1.0',
            'arch':'all',
            }
        gam_package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual(
            [bar_package, gam_package],
            foo_package.find_unsafely_colocated_packages())
        self.assertEqual(
            [bar_package, foo_package],
            gam_package.find_unsafely_colocated_packages())
        self.assertEqual(
            [foo_package, gam_package],
            bar_package.find_unsafely_colocated_packages())

    def test_find_unsafely_colocated_packages_candidates_partial_diverted(self):
        """Only diverting some files leaves the packages still unsafe."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # cheap way to insert a package.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/etc/diverted',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            'diversions':['/etc/diverted'],
            }
        foo_package = PackageVersion.InsertFromInfo(info, store)
        # a alternate package for this one, on the same arch.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/etc/diverted',
                '/usr/share/doc/bar/copyright',
                ],
            'name':'bar',
            'version':'0.2',
            'arch':'sparc',
            }
        bar_package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual(
            [bar_package],
            foo_package.find_unsafely_colocated_packages())
        self.assertEqual(
            [foo_package],
            bar_package.find_unsafely_colocated_packages())

    def test_find_unsafely_colocated_packages_candidates_full_diverted(self):
        """Diverting all the common filess leaves the packages safe."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # cheap way to insert a package.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/etc/diverted',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            'diversions':['/etc/diverted'],
            }
        foo_package = PackageVersion.InsertFromInfo(info, store)
        # a alternate package for this one, on the same arch.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/etc/diverted',
                '/usr/share/doc/bar/copyright',
                ],
            'name':'bar',
            'version':'0.2',
            'arch':'sparc',
            'diversions':['/etc/foo'],
            }
        bar_package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual(
            [],
            foo_package.find_unsafely_colocated_packages())
        self.assertEqual(
            [],
            bar_package.find_unsafely_colocated_packages())

    def test_find_common_files_none(self):
        """No files in common, nothing to return."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            }
        foo_package = PackageVersion.InsertFromInfo(info, store)
        # and a second non-overlapping package.
        info = {
            'controltext':'',
            'files':[
                '/etc/bar',
                '/usr/share/doc/bar/copyright',
                ],
            'name':'bar',
            'version':'0.2',
            'arch':'sparc',
            }
        bar_package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual([], foo_package.find_common_files(bar_package))

    def test_find_common_files(self):
        """With files in common, they are returned in sorted order."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/bin/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            }
        foo_package = PackageVersion.InsertFromInfo(info, store)
        # and a second overlapping package.
        info = {
            'controltext':'',
            'files':[
                '/etc/foo',
                '/usr/bin/foo',
                '/usr/share/doc/bar/copyright',
                ],
            'name':'bar',
            'version':'0.2',
            'arch':'sparc',
            }
        bar_package = PackageVersion.InsertFromInfo(info, store)
        self.assertEqual([Filepath('/etc/foo'), Filepath('/usr/bin/foo')],
            foo_package.find_common_files(bar_package))

    def test_get_info(self):
        """The get_info function returns a dict for MissingConflictFinder."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        info = {
            'controltext':'Version: foo\n',
            'files':[
                '/etc/foo',
                '/usr/bin/foo',
                '/usr/share/doc/foo/copyright',
                ],
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            }
        package = PackageVersion.InsertFromInfo(info, store)
        # default condition: no conflicts/provides/replaces
        self.assertEqual({
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            'conflicts':'',
            'provides':'',
            'replaces':'',
            }, package.get_info())
        # conflicts
        package.controltext = (
            "Version: foo\n"
            "Conflicts: bar\n")
        self.assertEqual({
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            'conflicts':'bar',
            'provides':'',
            'replaces':'',
            }, package.get_info())
        # provides
        package.controltext = (
            "Version: foo\n"
            "provides: mta\n")
        self.assertEqual({
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            'conflicts':'',
            'provides':'mta',
            'replaces':'',
            }, package.get_info())
        # replaces
        package.controltext = (
            "Version: foo\n"
            "replaces: something\n")
        self.assertEqual({
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            'conflicts':'',
            'provides':'',
            'replaces':'something',
            }, package.get_info())
        # all
        package.controltext = (
            "Version: foo\n"
            "conflicts: something << 1.4\n"
            "provides: mta\n"
            "replaces: something\n")
        self.assertEqual({
            'name':'foo',
            'version':'0.1-0ubuntu2',
            'arch':'sparc',
            'conflicts':'something << 1.4',
            'provides':'mta',
            'replaces':'something',
            }, package.get_info())

    def test_To_analyze(self):
        """Find the ids of versions that need analysis."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # add foo-1.2-all and quux-1.3-i386
        all = PackageArchitecture.Ensure('all', store)
        i386 = PackageArchitecture.Ensure('i386', store)
        foo = PackageName.Ensure('foo', store)
        quux = PackageName.Ensure('quux', store)
        store.add(PackageVersion(foo, '1.2', all, '', analyzed_level=1))
        quux_version = PackageVersion(quux, '1.3', i386, '')
        store.add(quux_version)
        store.flush()
        expected_result = set([quux_version.id])
        self.assertEqual(expected_result,
            set(PackageVersion.To_analyze(store)))
        self.assertEqual(set(),
            set(PackageVersion.To_analyze(store, [('foo', '1.2', 'all')])))
        self.assertEqual(expected_result,
            set(PackageVersion.To_analyze(store,
                [('foo', '1.2', 'all'), ('quux', '1.3', 'i386')])))


class TestFilepaths(TestCase):

    def test___eq__(self):
        self.assertNotEqual(None, Filepath('foo'))
        self.assertNotEqual(Filepath('bar'),
            Filepath('foo'))
        self.assertEqual(Filepath('foo'),
            Filepath('foo'))

    def test___lt__(self):
        self.assertTrue(Filepath('') < Filepath('abc'))
        self.assertFalse(Filepath('abc') < Filepath(''))
        self.assertTrue(Filepath('/etc') < Filepath('/usr'))
        self.assertFalse(Filepath('/usr') < Filepath('/etc'))

    def test_initialize(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # a new db should have no known paths
        self.assertEqual([], list(store.find(Filepath)))

    def test_fields(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # test we can insert objects with the expected fields,
        # and get them back.
        _usr_bin_mail = Filepath('/usr/bin/mail')
        store.add(_usr_bin_mail)
        # we should now have it safely in the database. committing 
        # and doing a garbage collect should let us get at it 
        # uncached.
        del _usr_bin_mail
        gc.collect()
        store.commit()
        self.assertEqual([Filepath('/usr/bin/mail')],
            list(store.find(Filepath)))
        # we should also be unable to add duplicates.
        new_usr_bin_mail = Filepath('/usr/bin/mail')
        store.add(new_usr_bin_mail)
        self.assertRaises(IntegrityError, store.commit)
        # having made it fail, rollback.
        store.rollback()
        self.assertEqual([Filepath('/usr/bin/mail')],
            list(store.find(Filepath)))
        # we should be able to get to the binpackages for a file.
        foo = PackageVersion(PackageName('foo'), '3', PackageArchitecture('all'),
            '')
        bar = PackageVersion(PackageName('bar'), '1', PackageArchitecture('i386'),
            '')
        store.add(foo)
        store.add(bar)
        path = list(store.find(Filepath))[0]
        path.binpackages.add(foo)
        path.binpackages.add(bar)
        self.assertEqual(set([foo, bar]), set(path.binpackages))

    def test_Ensure(self):
        """Ensure should provide an easy api for working with Filepaths."""
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # create a Filepath in the store.
        _usr_bin_mail = Filepath.Ensure('/usr/bin/mail', store)
        store.commit()
        self.assertEqual([Filepath('/usr/bin/mail')],
            list(store.find(Filepath)))
        # and reuse an existing one if possible.
        a = Filepath.Ensure('/foo', store)
        b = Filepath.Ensure('/foo', store)
        self.assertTrue(a is b)
        

class TestPackageFiles(TestCase):

    def test___eq__(self):
        self.assertNotEqual(None, Filepath('foo'))
        self.assertNotEqual(Filepath('bar'),
            Filepath('foo'))
        self.assertEqual(Filepath('foo'),
            Filepath('foo'))

    def test_initialize(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # a new db should have no known paths
        self.assertEqual([], list(store.find(Filepath)))

    def test_fields(self):
        # an in memory test db
        store = ArchiveDatabase.initialize('sqlite:')
        # test we can insert objects with the expected fields,
        # and get them back.
        _usr_bin_mail = Filepath('/usr/bin/mail')
        store.add(_usr_bin_mail)
        # we should now have it safely in the database. committing 
        # and doing a garbage collect should let us get at it 
        # uncached.
        del _usr_bin_mail
        gc.collect()
        store.commit()
        self.assertEqual([Filepath('/usr/bin/mail')],
            list(store.find(Filepath)))
        # we should also be unable to add duplicates.
        new_usr_bin_mail = Filepath('/usr/bin/mail')
        store.add(new_usr_bin_mail)
        self.assertRaises(IntegrityError, store.commit)
        # having made it fail, rollback.
        store.rollback()
        self.assertEqual([Filepath('/usr/bin/mail')],
            list(store.find(Filepath)))


### TODO ###: investiage the del / commit / gc interaction.
### TODO ###: indirect reference - link tables - need to be default
### constructable, it should say this.
### TODO ###: two indirect links to each other: how to do without 'poke
### at the object idiom'.
### TODO XXX: a result set returning a tuple returns None rather than
### (None, None, None) for one().
### TODO XXX: selecting a 100K object list, then committing one object takes 10 seconds to do the commit.

# planned transform:
#   instead of 'add -> possible conflicts'
#   make packageversion->files cheap to lookup
#   then, find unanalyzed packages, query for 'files in packages with different
#   names and either the same arch or the inverse-around-all arch'.
#   that gives us the smallest relevant output.
#
# things that we do:
# collection of filepath : package versions. 
#   add_files -> returns a dict with keys of package combinations, and paths
#                that conflicted. we will want to not report packages with
#                the same name at all.
#
# we determine new packages by comparing the ones we've already inserted
#   that is, the package is not in packageversions.
#   for speed, we want a set operator here. One way would be to insert
#   all the new packages into a temp table, then ask. Another would be
#   to insert all the packages and then query for those without file data,
#   or without a 'checked' flag. Alternatively we can say we do X forms
#   of analysis and search for that.
